/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_LinkRx.cpp

COPYRIGHT (C) 2002   Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
                     Cristian Giussani (cgiussani@fastflow.it)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef LINKRX_CPP
#define LINKRX_CPP

using namespace std;

#include <DCDT_LinkRx.h>
#include <DCDT_Finder.h>

static void InstallSignalHandler ( void )
{
  struct sigaction sa;
  __sigset_t sigset;
  
  // Ignore SIGPIPE from sockets
  sa.sa_handler = SIG_IGN;
  memset(&sigset,0,sizeof(sigset));;
  sa.sa_mask = sigset;
  sa.sa_flags = 0;
  sa.sa_restorer = NULL;
  if ( sigaction(SIGPIPE,&sa,NULL) != 0 ) {
    printf ( "Error: Unable to ignore SIGPIPE\n" );
    exit ( 1 );
  }
}

DCDT_LinkRx::DCDT_LinkRx(DCDT_Agora *agora, int id, int r_id) : DCDT_Member(agora)
{
  linkID = id;
  remoteID = r_id;
  status = L_WAITING;
  restart_counter = 0;
  for (int i=0; i<NUMINT_MSGTYPE; i++)
    mymask[i] = 0;
  localCD = NULL;
  remoteCD = NULL;
  msgrcv = NULL;
  LinkTxPtr = NULL;
  InstallSignalHandler();
}

void DCDT_LinkRx::Init()
{
}

/*! \brief Prepare the data structures and open the local channel part
 *
 *  l_cd must be specified, at exit every null field will contains values defined by the kernel.
 * \return Pointer to the communications data for the channel
 */ 
CommData* DCDT_LinkRx::Prepare(int stat, CommData *l_cd, CommData *r_cd)
{
  TRC_ENTER( DCDT_TRC_LINKRX, TRC1, ("Status %d - LocalCD %p - RemoteCD %p", stat, l_cd, r_cd ));
  channel = l_cd->CreateChannel(myAgoraID);
  try {
    channel->Prepare(l_cd, r_cd);
  } catch ( ChannelError e ) {
    cout << "channel->Prepare ChannelError " << e.errval << "\n";
    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("LinkRx - Channel->Prepare ChannelError %d\n", e.errval));
  }
  
  try {
    channel->Open(creation_status = status = stat);
  } catch ( ChannelError e ) {
    cout << "channel->Open ChannelError " << e.errval << "\n";
    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("LinkRx - Channel->Open ChannelError %d\n", e.errval));
  }
  
  channel->SetTimers();
  CommData *ret;
  try {
    ret = channel->GetCommData();
  } catch ( ChannelError e ) {
    cout << "channel->GetCommData ChannelError " << e.errval << "\n";
    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("LinkRx - Channel->GetCommData ChannelError %d\n", e.errval));
  }
	     
  TRC_RETURN( DCDT_TRC_LINKRX, TRC1,("Ret CommData %p", ret), ret );
  return ret;
}


/*! \brief FSM for the LinkRx member.
 *
 *  - L_WAITING: Wait for a connection and an MT_CONNECT message
 *  - L_CONNECTING: Connect to a remote link and send an MT_CONNECT message
 *  - L_RESETTOCONN: Stop and restart the channel
 *  - L_WORKING: Receive messages and if needed share it with local members
 *  - L_STOPPING: End all the communications 
 *  - L_LOST: The communication is lost, if the channel remains in this state for a certain amount of time it can be deleted and his resources can be released.
 *
 *  In any state if a timeout occurs the state is always set to L_STOPPING.
 */
void DCDT_LinkRx::DoYourJob(int par)
{
  TRC_ENTER( DCDT_TRC_LINKRX, TRC1, ("Status %d", status ));
  try {
    switch (status) {
      
    case L_WAITING:
      TRC_PRINT( DCDT_TRC_COMM, TRC1, ("LinkRx %i channel waiting", linkID));
      try {
        channel->WaitConn();
      } catch (TimeOut t) {
        if (status != L_RESETTOCONN) {
	  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Exception - L_WAITING & status != L_RESETTOCONN"));
          throw;
	}
        else break;
      }
      
      // If the status is L_RESETTOCONN I don't care
      // I have to establish a link
      status = L_WORKING;
      channel->SetTimers();
      channel->ReceiveTimerOff();
      if (LinkTxPtr) {
        LinkTxPtr->force_mask=true; 
        LinkTxPtr->Restart(channel);
	TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Restarting LinkTx"));
      }
      else {
        LinkTxPtr = new DCDT_LinkTx(myAgora, linkID, remoteID, this);
        myAgora->AddSysMember(LinkTxPtr, Profile);
        LinkTxPtr->Prepare(channel);
	TRC_PRINT( DCDT_TRC_COMM, TRC1, ("LinkTx created!"));
        myAgora->ActivateMember(LinkTxPtr);
      }
      break;
      
    case L_CONNECTING:
      TRC_PRINT( DCDT_TRC_COMM, TRC1, ("LinkRx %i Wow, we're connecting!!!", linkID));
      channel->StartConn();
      status = L_WORKING;
      channel->SetTimers();
      channel->ReceiveTimerOff();
      if (LinkTxPtr) {
        LinkTxPtr->force_mask=true;
        LinkTxPtr->Restart(channel);
	TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Restarting LinkTx"));
      }
      else {
        LinkTxPtr = new DCDT_LinkTx(myAgora, linkID, remoteID, this);
        myAgora->AddSysMember(LinkTxPtr, Profile);
        LinkTxPtr->Prepare(channel);
	TRC_PRINT( DCDT_TRC_COMM, TRC1, ("LinkTx created!"));
        myAgora->ActivateMember(LinkTxPtr);
      }
      break;
      
    case L_RESETTOCONN:
      TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Reset LinkRx %i to L_CONNECTING", linkID));
      channel->Stop();
      channel->Dispose();
      Restart(L_CONNECTING, localCD, remoteCD);
      break;
      
    case L_WORKING:
      TRC_PRINT( DCDT_TRC_COMM, TRC1, ("LinkRx %i WORKING", linkID));
      
      // Receiving
      while ((msgrcv = channel->Receive())) {
	TRC_PRINT( DCDT_TRC_COMM, TRC1, ("LinkRx %i received remote msg type %i", linkID, msgrcv->ReadType()));
        if ((msgrcv->ReadType()) == MT_SUBSUPD) {
          LinkTxPtr->UnSubscribeAll();
          subs_mask_rcv = (unsigned int*)msgrcv->GetPayload();
          LinkTxPtr->Subscribe(subs_mask_rcv);
          delete msgrcv;
        }
        else if ((msgrcv->ReadType()) == MT_KEEPALIVE) {
          delete msgrcv;
        }
        else {
          LinkTxPtr->ShareMsg(msgrcv);
        }
      }
      break;
      
    case L_STOPPING:
      TRC_PRINT( DCDT_TRC_COMM, TRC1, ("LinkRx %i STOPPING", linkID));
      channel->Stop();
      channel->Dispose();
      lost_counter = 0;
      status = L_LOST;
      
    case L_LOST:
      TRC_PRINT( DCDT_TRC_COMM, TRC1, ("LinkRx %i LOST (%d)", linkID, lost_counter));
      if ((lost_counter++) == LINK_LOST_NUM) {
        TRC_PRINT( DCDT_TRC_COMM, TRC1, ("LinkRx %i KILLED", linkID));
        Close();
        if ( LinkTxPtr )
	  myAgora->RemoveMember(LinkTxPtr);
        status = L_KILL;
        //myFinder->CloseLink(linkID);
        Exit();
        break;
      }
      Delay(LINK_LOST_TIMER);
      break;
      
    case L_KILL:
      TRC_PRINT( DCDT_TRC_COMM, TRC1, ("LinkRx %i KILL", linkID));
      Delay(LINK_LOST_TIMER);
      break;
      
    }
  } catch(TimeOut e) {
    cout << "\nTimeout on linkrx " << linkID << "\nConnection lost";
    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("TimeOut on LinkRx %i", linkID));
    status = L_STOPPING;
    if ( LinkTxPtr ) {
      LinkTxPtr->UnSubscribeAll();
      LinkTxPtr->ChangeStatus(L_LOST);
      LinkTxPtr->force_mask=true;
    }
  } catch(ConnError e) {
    cout << "\nError on linkrx " << linkID << "\nerrno=" << e.errval << "\nConnection lost";
    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Error on LinkRx %i = %d (1)", linkID, e.errval));
    status = L_STOPPING;
    if ( LinkTxPtr ) {
      LinkTxPtr->UnSubscribeAll();
      LinkTxPtr->ChangeStatus(L_LOST);
      LinkTxPtr->force_mask=true;
    }
  } catch(ChannelError e) {
    cout << "\nError on linkrx " << linkID << "\nerrno=" << e.errval << "\nConnection lost";
    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Error on LinkRx %i = %d (2)", linkID, e.errval));
    status = L_STOPPING;
    if ( LinkTxPtr ) {
      LinkTxPtr->UnSubscribeAll();
      LinkTxPtr->ChangeStatus(L_LOST);
      LinkTxPtr->force_mask=true;
    }
  } catch(Exception e) {
    cout << "\nException! Closing linkrx " << linkID << "\n";
    TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Exception on LinkRx %i", linkID));
    Close();
    if ( LinkTxPtr )
      myAgora->RemoveMember(LinkTxPtr);
    //myFinder->CloseLink(linkID);
    status = L_KILL;
  }

  TRC_VOID_RETURN( DCDT_TRC_LINKRX, TRC1);
}


CommData* DCDT_LinkRx::Restart(int re_status, CommData *l_cd, CommData *r_cd)
{
  cout << "\nRestarting LINKRX " << linkID;
  TRC_PRINT( DCDT_TRC_COMM, TRC1, ("Restarting LinkRx %i", linkID));
  channel->Prepare(l_cd, r_cd);
  channel->Reopen(creation_status = status = re_status);
  l_cd = channel->GetCommData();
  channel->SetTimers();
  return l_cd;
}

CommData* DCDT_LinkRx::GetCommData()
{
  return channel->GetCommData();
}

void DCDT_LinkRx::ResetToConnecting(CommData *l_cd, CommData *r_cd)
{
  SetCDs(l_cd, r_cd);
  status = L_RESETTOCONN;
}

void DCDT_LinkRx::SetCDs(CommData *l_cd, CommData *r_cd)
{
  if (localCD)
    delete localCD;
  localCD = l_cd;
  if (remoteCD)
    delete remoteCD;
  remoteCD = r_cd;
}

void DCDT_LinkRx::Close()
{
  channel->Close();
}

#endif
