/****************************************************************************
dEVICE cOMMUNITIES dEVELOPMENT tOOLKIT 

DCDT_PostOffice.h

COPYRIGHT (C) 2002  Paolo Meriggi (meriggi@ing.unibs.it)
                    Alessandro Mazzini (mazzini@airlab.elet.polimi.it)
                    Cristian Giussani (cgiussani@fastflow.it)


This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.

****************************************************************************/

#ifndef DCDT_POSTOFFICE_SLWBU_H
#define DCDT_POSTOFFICE_SLWBU_H

/*! \brief Informations about the single message stored.
 *
 *  When an elemet was deleted it checks if the message was read by all the members, if this is the last member the message is deleted, otherwise the readers counter is decremented.
 */
class DCDT_MsgListElem_SLWBU
{
 public:
  DCDT_MsgListElem_SLWBU ()
    {
      ToBeDeleted = false;
      MsgPtr = NULL;
      prev = NULL;
      next = NULL;

      for(int i = 0; i < NUMINT_MEMBERS; i++) {
	readers[i] = 0;
      }
    };
  
  ~DCDT_MsgListElem_SLWBU ()
    { 
      // If this is the last reader we have to delete the message.
      // Oterwise we have to decrement the readers counter
      TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("Msg %p Readers %d", MsgPtr, MsgPtr->Get_Readers_Counter()));
      if (MsgPtr->Get_Readers_Counter() == 1) {
        TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("Freeing Msg %p", MsgPtr));
	delete MsgPtr;
      }
      else
	MsgPtr->Decrement_Readers_Counter(); 
    };
  
  struct timeval insert_time;  
  
  bool ToBeDeleted;

  /// Flag: true -> message requested by a local user member ( member profile = MPM_USER )  
  bool local;
  
  int LocalPublisherID;
  
  DCDT_MsgListElem_SLWBU *next;
  DCDT_MsgListElem_SLWBU *prev;
    
  /// Array of bit reporting members subscribed to this message type
  unsigned int readers[NUMINT_MEMBERS];
  
  /// Type of the message
  unsigned int type_mask[NUMINT_MSGTYPE];

  /// Message
  DCDT_Msg *MsgPtr;
  
  /** Mutex used when modifing the mask */
  DCDT_Mutex mtxmask;  
};



// Structure for the registered members
typedef struct _MemberTableElem_SLWBU
{
  DCDT_Mutex mtx;
  
  // Flag: reporting whether or not this slot is hosting a member
  bool alive;
  
  // Message types that this members is subscribed to
  // unsigned int SubscribedMsg[NUMINT_MSGTYPE];
  unsigned int LocalSubscribedMsg[NUMINT_MSGTYPE];
  unsigned int RemoteSubscribedMsg[NUMINT_MSGTYPE];

  // Flag: TRUE - the susbscription mask was updated
  bool update;
  
  // Pointers to the MemberTable when every single member has a buffer
  // Allow the buffer side

  DCDT_MsgListElem_SLWBU *HeadBuffer;
  DCDT_MsgListElem_SLWBU *TailBuffer;
  
  DCDT_Condition condition_var;

  //Flag: true if a member is waiting for a message
  bool waiting;
}
MemberTableElem_SLWBU;

/*! \brief Data structures and method to subscribe, insert, delete and read messages.
 *
 *  NOTE: This class isn't a member of the Agor, the real member that uses this class is DCDT_MsgManager.
 */
class DCDT_PostOffice_SLWBU
{
  friend class DCDT_MsgManager;
  friend class DCDT_Agora;
  friend class DCDT_Member;

 public:
    
  DCDT_PostOffice_SLWBU (DCDT_Agora *agora)
    {
      myAgora = agora;
      NumOfMsgPresent = 0;

      // MemberTable initialization
      for (int i = 0; i < MAX_MEMBERS; i++)
	{
	  MemberTable[i].alive = false;
	  MemberTable[i].update = false;
          MemberTable[i].waiting = false;

	  MemberTable[i].TailBuffer = NULL;
	  MemberTable[i].HeadBuffer = NULL;
	  
	  for (int j = 0; j < NUMINT_MSGTYPE; j++) {
	    // MemberTable[i].SubscribedMsg[j] = 0;
	    MemberTable[i].LocalSubscribedMsg[j] = 0;
	    MemberTable[i].RemoteSubscribedMsg[j] = 0;
	  }
	}
      
      for (int i = 0; i < MAX_MSG_TYPE; i++)
	{
	  for (int j = 0; j < NUMINT_MEMBERS; j++) 
	    {
	      SubscribedTable[i].SubscribersIfLocal[j] = 0;
	      SubscribedTable[i].SubscribersIfRemote[j] = 0;
	    }
	}
      printf ("Postoffice created\n");
    };
  

  ~DCDT_PostOffice_SLWBU ()
    {
      // TODO: clean of MsgList
      for ( int i=0; i < MAX_MEMBERS; i++ )
	{
	  if ( MemberTable[i].alive = true ) {
	    DCDT_MsgListElem_SLWBU *Elem=MemberTable[i].HeadBuffer;
	    while ( Elem != NULL ) {
	      Elem=RemoveBufferElement ( i, Elem, 1 ); 
	    }
	  }
	}
    };

  inline void AddMember(int Mem_ID);
  inline void AddSysMember(int Mem_ID);
  
  inline void DeleteMemberTableEntry(int Mem_ID );
  
  inline void AddMsg (DCDT_Msg * MP, int Mem_ID, DCDT_RequestType Type);
  
  inline DCDT_Msg *ReadNextMsg (int MemID, bool wait);
  inline DCDT_Msg *ReadNextMsg (int MemID, int type, bool wait );

  inline DCDT_Msg *ReadLastMsg ( int MemID, int type, bool wait );

  inline void CleanupList ();
  
  inline void SubscribeMsgTypeID (int Member_ID, int MsgType, DCDT_RequestType ReqType = DCDT_ALL_MSG );
  inline void UnSubscribeMsgTypeID ( int Member_ID, int MsgType, DCDT_RequestType ReqType = DCDT_ALL_MSG );
  inline void UnSubscribeAll( int Member_ID );
  
  inline void UpdateSubscribersMask (int MemID);
  inline bool ReadSubscribedMsgMask(unsigned int *mask, int MemID);
  inline bool UpdateSubs(int MemID);
    
 private:
  
  inline void SetReaders ( DCDT_MsgListElem_SLWBU * Elem, int num_type); 
  
  inline DCDT_MsgListElem_SLWBU *RemoveBufferElement( int MemID, DCDT_MsgListElem_SLWBU *elem, int dir);
  
  inline int MatchMsgTypeMask (int Member_ID, DCDT_MsgListElem_SLWBU *);
  inline int MatchReadersMask (int Member_ID, DCDT_MsgListElem_SLWBU *);
  
  inline void SetType (DCDT_MsgListElem_SLWBU * Elem, int type);

  // Functions used during the test phase
#ifdef TEST_PART
  inline void InsertListTime (int MemID, int TypeOp);
  inline void VisualList ();
  inline void SaveList (char* nomefile);
#else
#define InsertListTime(A,B)
#define VisualList()
#define SaveList(A)
#endif
    
#ifdef USE_TIMEOUT
  inline bool CheckTimeoutElem(DCDT_MsgListElem_SLWBU *); 
#else
#define CheckTimeoutElem(A) false
#endif

  inline int GetTypeFromMask (DCDT_MsgListElem_SLWBU * Elem);
  
  /*! \brief For each message type we have the corresponding member mask.
   *
   *  A member mask is a sequence of bit to repesent all the members, if a bit is set to 1 the corresponding member is subscribed to this type of messages.
   *  Ex: 00010001 -> Members 1 and 5 are subscribed to this message type.
   *  The member mask is composed of two parts local and remote, the local one was used when a message was locally originated or was received from a BRIDGE member; the remote one was used when the message was received from a STATIC or LINK member.
   */
  SubsMsgTypeElem SubscribedTable[MAX_MSG_TYPE];
  
  /// Mutex to access SubscribedMsgMask
  DCDT_Mutex subscribed_mtx;

  /// Remote message types requested by local members used by LinkTx members
  unsigned int SubscribedMsgMask[NUMINT_MSGTYPE];
  
  /// ???
  unsigned int LocalSubscribedMsgMask[NUMINT_MSGTYPE];
  unsigned int RemoteSubscribedMsgMask[NUMINT_MSGTYPE];

  // Mutex to access UpdateMask_flag and MemberTable[i].update
  DCDT_Mutex UpdateMask_mtx;

  /// Flag: true when SubscribedMsgMask was updated
  bool UpdateMask_flag;
  
  MemberTableElem_SLWBU MemberTable[MAX_MEMBERS];
  
  int NumOfMsgPresent;

  DCDT_Agora *myAgora;
};




inline void DCDT_PostOffice_SLWBU::AddMember(int Mem_ID) {
  MemberTable[Mem_ID].alive = true;
};

inline void DCDT_PostOffice_SLWBU::AddSysMember(int Mem_ID) {
  MemberTable[Mem_ID].alive = true;
};

inline void DCDT_PostOffice_SLWBU::DeleteMemberTableEntry(int Mem_ID){
  
  // TODO
  MemberTable[Mem_ID].alive = false;
  
};

/*! \brief Add a new message to the PostOffice queue.
 *
 *  When the member \aMem_ID wants to insert a new message the same member checks the subscription table and when a member is subscribed to those messages copy the message into the member message queue; this function also updates the message with the number of members subscribed to this message type.
 */
inline void DCDT_PostOffice_SLWBU::AddMsg (DCDT_Msg *MP, int Mem_ID, DCDT_RequestType Type)
{
  int msg_type;
  int temp;
  int MemberID;
  bool used=false;

  InsertListTime(Mem_ID,WRITEINIT);
  DCDT_MsgListElem_SLWBU *newMsgListElem;
  
  TRC_ENTER ( DCDT_TRC_POSTOFFICE, TRC1, ( "MemID: %d", Mem_ID));
  
  msg_type = MP->ReadType ();
  if ( msg_type > MAX_MSG_TYPE ) {
    // Quando riprende un link interno da un'agora' fermata e fatta ripartire
    return;
  }
  
  // Check if someone is subscribed to this message
  // Put the message in each member's local queue
  MP->Lock();
  MemberID=0;
  for (int i=0; i<NUMINT_MEMBERS;i++) { 
    for (int t=1;t<=SIZEOF_INTEGER;t++) { 
      MemberID++;
      
      switch ( Type ) {
      case DCDT_LOCAL_MSG:
	temp = SubscribedTable[msg_type].SubscribersIfLocal[i];
	temp = temp << (SIZEOF_INTEGER - t);
	temp = temp >> (SIZEOF_INTEGER-1);
	break;
	
      case DCDT_REMOTE_MSG:
	temp = SubscribedTable[msg_type].SubscribersIfRemote[i];
	temp = temp << (SIZEOF_INTEGER - t);
	temp = temp >> (SIZEOF_INTEGER-1); 
	break;
	
      default:
	printf ( "Unknown type of message ( LOCAL/REMOTE ) %d\n", Type );
	break;
      }
      
      if (temp && (MemberID != Mem_ID)) {
        used=true;
	MemberTableElem_SLWBU *member=&(MemberTable[MemberID]);

	// We have a member interested so we have to allocate a new element for the local message queue and copy the message pointer
	MP->Increment_Readers_Counter(false);
	newMsgListElem = new (DCDT_MsgListElem_SLWBU);
	newMsgListElem->MsgPtr = MP; 
	newMsgListElem->next = NULL;
	TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("AddMsg %p to member %d (%d)", MP, MemberID, newMsgListElem->ToBeDeleted ));
	newMsgListElem->ToBeDeleted=false;
	
	gettimeofday(&(newMsgListElem->insert_time), NULL);
	SetType (newMsgListElem, msg_type);
	InsertListTime(Mem_ID,WRITELOCKINIT);
	
	member->mtx.lock();
	member->condition_var.lock();
	
	InsertListTime(Mem_ID,WRITELOCKEND);
	
	if (member->HeadBuffer) {
	  member->TailBuffer->next = newMsgListElem;
	  newMsgListElem->prev = member->TailBuffer;
	}
	else
	  member->HeadBuffer = newMsgListElem;
	
	member->TailBuffer = newMsgListElem;
	
	// If someone is waiting send a signal
	if ( member->waiting ) {
	  member->waiting = false;
	  member->condition_var.signal();
	}
	
	NumOfMsgPresent++;
	
	member->condition_var.unlock();
	member->mtx.unlock();
      }
    }
  }
  
  MP->Unlock();

  // If no member is interested  in this message delete it
  if ( used == false ) {
     delete MP;
  }

  InsertListTime(Mem_ID,WRITEEND);
  
  TRC_VOID_RETURN ( DCDT_TRC_POSTOFFICE, TRC1 );
};

/*! \brief Read next message for the member \a MemID.
 *
 *  Check if there is an unread message in the member queue; if we found one we have to mark it as deletable and return a pointer to it.
 *  If, during the queue slide, we find a message marked as deletable we have to delete it and go on.
 *  \param MemID - Member identifier for which we want the next message
 *  \return NULL - No new messages available.
 *  \return !=NULL - Next message available.
 */
inline DCDT_Msg *DCDT_PostOffice_SLWBU::ReadNextMsg (int MemID, bool wait)
{
  DCDT_MsgListElem_SLWBU *Elem;

  TRC_ENTER ( DCDT_TRC_POSTOFFICE, TRC1, ("MemID: %d", MemID) );

  InsertListTime(MemID,READINIT);
  InsertListTime(MemID,READLOCKINIT);
  MemberTable[MemID].mtx.lock(); 
  InsertListTime(MemID,READLOCKEND);

  Elem = MemberTable[MemID].HeadBuffer;
  while ( 1 ) {

    if ( Elem == NULL ) {
      // There isn't any new message
      MemberTable[MemID].mtx.unlock();
      InsertListTime ( MemID, READDONEEND );

      if ( wait == true ) {
	// Wait a signal from another member when a new message will be available
	InsertListTime(MemID,READLOCKINIT);
	MemberTable[MemID].condition_var.lock(); 
	InsertListTime(MemID,READLOCKEND); 
	InsertListTime(MemID,WAITINIT);
	MemberTable[MemID].waiting = true;
	MemberTable[MemID].condition_var.wait();
	InsertListTime(MemID,WAITEND);
	MemberTable[MemID].condition_var.unlock();
	
	MemberTable[MemID].mtx.lock();
	InsertListTime(MemID,READLOCKEND);

	// Rescan the message queue
	Elem = MemberTable[MemID].HeadBuffer;
	continue;
      }
      else {
	InsertListTime(MemID,READEND);
	TRC_RETURN ( DCDT_TRC_POSTOFFICE, TRC1, ("Msg NULL"), NULL );
	return ((DCDT_Msg *) NULL);
      }
    }

    // This message has to be deleted ?
    if (Elem->ToBeDeleted || CheckTimeoutElem(Elem)) {

      // Yes, deleted it and check the next message
      InsertListTime(MemID,DELETEINIT);
      Elem=RemoveBufferElement(MemID, Elem, 1 ); /* Errore */
      InsertListTime(MemID,DELETEEND);
    }
    else {
      Elem->ToBeDeleted = true;
      MemberTable[MemID].mtx.unlock();
      InsertListTime(MemID,READDONEEND);
      
      TRC_RETURN ( DCDT_TRC_POSTOFFICE, TRC1, ("Msg %p", Elem->MsgPtr), Elem->MsgPtr );
      return (Elem->MsgPtr);
    }
  }
};

/*! \brief Read next message of type \a type for the member \a MemID.
 *
 *  Check if there is an unread message of type \a in the member queue; if we found one we have to mark it as deletable and return a pointer.
 *  If, during the queue slide, we find a message marked as deletable we have to delete it and go on.
 *  \param MemID - Member ID
 *  \param type - Requested message type
 *  \return NULL - No new messages available.
 *  \return !=NULL - Next message available.
 */
inline DCDT_Msg *DCDT_PostOffice_SLWBU::ReadNextMsg ( int MemID, int type, bool wait )
{
  DCDT_MsgListElem_SLWBU *Elem;
  
  TRC_ENTER ( DCDT_TRC_POSTOFFICE, TRC1, ("MemberID:%d Type: %d", MemID, type ) );
  InsertListTime(MemID,READINIT);
  InsertListTime(MemID,READLOCKINIT);
  MemberTable[MemID].mtx.lock(); 
  InsertListTime(MemID,READLOCKEND);
  
  Elem = MemberTable[MemID].HeadBuffer;
  
  while ( 1 ) {
    
    if ( Elem == NULL ) {
      // There isn't any new message
      MemberTable[MemID].mtx.unlock();
      InsertListTime ( MemID, READDONEEND );
      
      if ( wait == true ) {
	// Wait a signal from another member when a new message will be available
	InsertListTime(MemID,READLOCKINIT);
	MemberTable[MemID].condition_var.lock(); 
	InsertListTime(MemID,READLOCKEND); 
	InsertListTime(MemID,WAITINIT);
	MemberTable[MemID].waiting = true;
	MemberTable[MemID].condition_var.wait();
	InsertListTime(MemID,WAITEND);
	MemberTable[MemID].condition_var.unlock();
	
	MemberTable[MemID].mtx.lock();
	InsertListTime(MemID,READLOCKEND);

	// Rescan the message queue
	Elem = MemberTable[MemID].HeadBuffer;
	continue;
      }
      else {
	InsertListTime(MemID,READEND);
	TRC_RETURN ( DCDT_TRC_POSTOFFICE, TRC1, ("Msg NULL"), NULL);
	return ((DCDT_Msg *) NULL);
      }
      
    }

    // Delete all the elements marked or too old
    if ( Elem->ToBeDeleted || CheckTimeoutElem( Elem ) ) {
      TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("Msg %p ToBeDeleted %d", Elem->MsgPtr, Elem->ToBeDeleted));
      InsertListTime( MemID, DELETEINIT );
      Elem=RemoveBufferElement( MemID, Elem, 1 );
      InsertListTime( MemID, DELETEEND );
      continue;
    }
    
    // If the mesage is of the requested type return it
    if (( type == MSG_ALL_TYPES ) || ( Elem->MsgPtr->ReadType() == type )) {
      Elem->ToBeDeleted = true;
      
      MemberTable[MemID].mtx.unlock();
      InsertListTime(MemID,READDONEEND);
      InsertListTime(MemID,READEND);

      TRC_RETURN ( DCDT_TRC_POSTOFFICE, TRC1, ("Msg %p", Elem->MsgPtr), Elem->MsgPtr );
      return (Elem->MsgPtr);
    }
    
    Elem = Elem->next;
  }
}

/*! \brief Read the last message of type \a type received by member \a MemID.
 *
 *  Check for the last message of type \a type in the member queue; if we found it we have to mark it as deletable and return a pointer.
 *  We scan the whole queue and delete all the old messages of type \a type
 *  \param MemID -
 *  \param type - Requested message type
 *  \return NULL - No new messages available.
 *  \return !=NULL - Next message available.
 */
inline DCDT_Msg *DCDT_PostOffice_SLWBU::ReadLastMsg ( int MemID, int type, bool wait )
{
  DCDT_MsgListElem_SLWBU *Elem;
  DCDT_Msg *msg=NULL;

  TRC_ENTER ( DCDT_TRC_POSTOFFICE, TRC1, ("MemberID:%d Type: %d", MemID, type ) );

  InsertListTime(MemID,READINIT);
  InsertListTime(MemID,READLOCKINIT);
  MemberTable[MemID].mtx.lock(); 
  InsertListTime(MemID,READLOCKEND);

  Elem = MemberTable[MemID].TailBuffer;
  
  // Scan the whole queue searching messages of the requested type
  while ( 1 ) {
    if ( Elem == NULL ) {
      if ( msg == NULL ) {
	// We haven't found any message
	MemberTable[MemID].mtx.unlock();
	InsertListTime(MemID,READDONEEND);

	if ( wait == true ) {
	  // Wait a signal from another member when a new message will be available
	  InsertListTime(MemID,READLOCKINIT);
	  MemberTable[MemID].condition_var.lock(); 
	  InsertListTime(MemID,READLOCKEND); 
	  InsertListTime(MemID,WAITINIT);
	  MemberTable[MemID].waiting = true;
	  MemberTable[MemID].condition_var.wait();
	  InsertListTime(MemID,WAITEND);
	  MemberTable[MemID].condition_var.unlock();
	  
	  MemberTable[MemID].mtx.lock();
	  InsertListTime(MemID,READLOCKEND);

	  // Rescan the message queue
	  Elem = MemberTable[MemID].TailBuffer;
	  continue;
	}
	else {
	  InsertListTime(MemID,READEND);
	  TRC_RETURN ( DCDT_TRC_POSTOFFICE, TRC1, ("Msg NULL"), NULL);
	  return ((DCDT_Msg *) NULL);
	}
      }
      else
	break;
    }

    // Delete all the elements marked or too old
    if ( Elem->ToBeDeleted || CheckTimeoutElem( Elem ) ) {
      InsertListTime( MemID, DELETEINIT );
      Elem=RemoveBufferElement( MemID, Elem, -1 );
      InsertListTime( MemID, DELETEEND );
      
      continue;
    }
    
    // If the message is of the requested type and is the last message received return it
    // if is an old message delete it
    if ( ( type == MSG_ALL_TYPES ) || ( Elem->MsgPtr->ReadType() == type ) ) {
      if ( msg == NULL ) {
	Elem->ToBeDeleted = true;
	
	InsertListTime(MemID,READDONEEND);
	
	msg = Elem->MsgPtr;
      }
      else {
	InsertListTime( MemID, DELETEINIT );
	TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("Delete Msg %p", Elem->MsgPtr));
	Elem = RemoveBufferElement( MemID, Elem, -1 );
	InsertListTime( MemID, DELETEEND );
	continue;
      }
    }
    
    Elem = Elem->prev;
  }

  InsertListTime(MemID,READDONEEND);
  MemberTable[MemID].mtx.unlock();

  // No new messages available
  InsertListTime(MemID,READEND);

  TRC_RETURN ( DCDT_TRC_POSTOFFICE, TRC1, ("Msg %p", msg), msg );
  return ((DCDT_Msg *) msg );
}

/*! \brief Delete the head element from \aMemId's local queue 
 *
 *  NB The message is deleted when needed with the destructor of ElemDeletable
 */
DCDT_MsgListElem_SLWBU *DCDT_PostOffice_SLWBU::RemoveBufferElement( int MemID, DCDT_MsgListElem_SLWBU *elem, int dir )
     
{
  DCDT_MsgListElem_SLWBU *ret=NULL;

  if ( elem != NULL ) {
    
    // Update the previous element
    if ( elem->prev != NULL )
      elem->prev->next = elem->next;

    // Update the next element
    if ( elem->next != NULL )
      elem->next->prev = elem->prev;
    
    if ( elem == MemberTable[MemID].HeadBuffer )
      MemberTable[MemID].HeadBuffer = elem->next;
    
    if ( elem == MemberTable[MemID].TailBuffer )
      MemberTable[MemID].TailBuffer = elem->prev;
    
    switch ( dir ) {
    case 1:
      ret = elem->next;
      break;
    case -1:
      ret = elem->prev;
      break;
    default:
      printf ( "Remove Err\n" );
      break;
    }
    TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("Deleting Msg %p", elem));
    delete elem;
    
    NumOfMsgPresent--;
  }
  
  return ret;
}

#ifdef USE_TIMEOUT
/*! \brief Check if the message pointed by  \a elem was expired
 */ 
inline bool DCDT_PostOffice_SLWBU::CheckTimeoutElem( DCDT_MsgListElem_SLWBU * elem)
{
  struct timeval now;
  DCDT_MsgListElem_SLWBU * elemtmp;
  pippo; /* ??? */
  elemtmp = NULL;

  if ( elem != elemtmp  ) 
    {
      gettimeofday(&now, NULL) ;     
      return ( DeltaSeconds( elem->insert_time, now) > TIMEOUT_MSG);
    }
  else
    return false;
};
#endif


inline void DCDT_PostOffice_SLWBU::CleanupList ()
{

}

/*! \brief Update the readers mask for the element \a Elem and message type \a num_type.
 */
inline void DCDT_PostOffice_SLWBU::SetReaders (DCDT_MsgListElem_SLWBU * Elem, int num_type)
{  
  int i;
  int PublisherID, offset, pos;
  unsigned int mask;


  TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("SetReaders"));
  
  if (Elem && ((num_type > 0) && (num_type <= MAX_MSG_TYPE))) {
    PublisherID = Elem->LocalPublisherID;
    
    offset = (PublisherID - 1) / SIZEOF_INTEGER;
    pos = PublisherID - (offset * SIZEOF_INTEGER);
    mask = ~(1 << (pos - 1));

    // Fill the readers mask 
    for (i = 0; i < NUMINT_MEMBERS; i++) {
      if (Elem->local || myAgora->CheckMemberProfile(PublisherID, MPM_BRIDGE))
	Elem->readers[i] = SubscribedTable[num_type].SubscribersIfLocal[i];
      else
	Elem->readers[i] = SubscribedTable[num_type].SubscribersIfRemote[i];
      
      
    };
    
    // Avoid self-reading
    Elem->readers[offset] &= mask;
  }
};

/* !\brief Subscribe member \a Member_ID to the messages of type \a MsgType.
 *
 */
inline void DCDT_PostOffice_SLWBU::SubscribeMsgTypeID (int Member_ID, int MsgType, DCDT_RequestType ReqType)
{
  int offset, pos;

  // Check if the message type is correct
  if ((MsgType > 0) && (MsgType <= MAX_MSG_TYPE)) {
    
    TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("Subscription request from Member %d to Msg Type %d ReqType %d", Member_ID, MsgType, ReqType));
    
    // Update the member subscription table
    // Don't need a lock because only the same member can update his subscription mask
    offset = (MsgType - 1) / SIZEOF_INTEGER;
    pos = MsgType - (offset * SIZEOF_INTEGER);

    if ( ReqType & DCDT_LOCAL_MSG )
      MemberTable[Member_ID].LocalSubscribedMsg[offset] |= (1 << (pos - 1));

    if ( ReqType & DCDT_REMOTE_MSG )
      MemberTable[Member_ID].RemoteSubscribedMsg[offset] |= (1 << (pos - 1));
    
#ifdef DCDT_USE_DEBUG
    printf ("MemberTable[%d].LocalSubscribedMsg=",Member_ID);
    for (int i = (NUMINT_MSGTYPE - 1); i >= 0; i--)
      printf ("[%x]", MemberTable[Member_ID].LocalSubscribedMsg[i]);
    printf ("\n");

    printf ("MemberTable[%d].RemoteSubscribedMsg=",Member_ID);
    for (int i = (NUMINT_MSGTYPE - 1); i >= 0; i--)
      printf ("[%x]", MemberTable[Member_ID].RemoteSubscribedMsg[i]);
    printf ("\n");
#endif
    
    // Update the message subscription table
    offset = (Member_ID - 1) / SIZEOF_INTEGER;
    pos = Member_ID - (offset * SIZEOF_INTEGER);
    
    SubscribedTable[MsgType].mtx.lock ();
    
    // Reading politics: 
    // - local members can read remote or locale msgs
    // - bridges read all msgs;
    // - remote members read local msgs

    // Local messages
    if ( ( myAgora->CheckMemberProfile(Member_ID, MPM_USER ) && ( ReqType & DCDT_LOCAL_MSG ) ) ||
	 ( myAgora->CheckMemberProfile(Member_ID, MPM_BRIDGE ) ) ||
	 ( myAgora->CheckMemberProfile(Member_ID, MPM_LINK ) ) )
      SubscribedTable[MsgType].SubscribersIfLocal[offset] |= (1 << (pos - 1));

    // Remote messages
    if ( ( myAgora->CheckMemberProfile(Member_ID, MPM_BRIDGE ) ) ||
	 ( ( ReqType & DCDT_REMOTE_MSG ) && (myAgora->CheckMemberProfile(Member_ID, MPM_USER)) ) )
      SubscribedTable[MsgType].SubscribersIfRemote[offset] |= (1 << (pos - 1));
    
    SubscribedTable[MsgType].mtx.unlock ();
    
    
#ifdef DCDT_USE_DEBUG
    printf ("SubscribedTable[%d].SubscribersIfLocal=",MsgType);
    for (int i = (NUMINT_MEMBERS - 1); i >= 0; i--)
      printf ("[%x]", SubscribedTable[MsgType].SubscribersIfLocal[i]);
    printf ( "\n" );

    printf("SubscribedTable[%d].SubscribersIfRemote=",MsgType);
    for (int i = (NUMINT_MEMBERS - 1); i>= 0; i--)
      printf ("[%x]", SubscribedTable[MsgType].SubscribersIfRemote[i]);
    printf ("\n");
#endif
    
    // Update the update flags
    UpdateMask_mtx.lock();
    for (int i=0; i<Member_ID; i++)
      MemberTable[i].update = true;
    for (int i=Member_ID+1; i<MAX_MEMBERS; i++)
      MemberTable[i].update = true;
    
    UpdateMask_flag = true;
    UpdateMask_mtx.unlock();
    
  } //end if valid msg type
};


/*! \brief Unsubscribe member \a Member_ID from the messages of type \a MsgType.
 *
 */
inline void DCDT_PostOffice_SLWBU::UnSubscribeMsgTypeID (int Member_ID, int MsgType, DCDT_RequestType ReqType )
{
  int  offset, pos;
  unsigned int mask;
  

  TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("UnSubscription request from Member %d to Msg Type %d ReqType %d", Member_ID, MsgType, ReqType));

  // Check if the message type is correct
  if ((MsgType > 0) && (MsgType <= MAX_MSG_TYPE)) {
    
    // Update member subscription table
    offset = (MsgType - 1) / SIZEOF_INTEGER;
    pos = MsgType - (offset * SIZEOF_INTEGER);
    mask = ~(1 << (pos - 1));
    
    if ( ReqType & DCDT_LOCAL_MSG )
      MemberTable[Member_ID].LocalSubscribedMsg[offset] &= mask;

    if ( ReqType & DCDT_REMOTE_MSG )
      MemberTable[Member_ID].RemoteSubscribedMsg[offset] &= mask;
    
#ifdef DCDT_USE_DEBUG
    printf ("MemberTable[%d].LocalSubscribedMsg=",Member_ID);
    
    for (int i = (NUMINT_MSGTYPE - 1); i >= 0; i--)
      printf ("[%x]", MemberTable[Member_ID].LocalSubscribedMsg[i]);
    printf ("\n");

    printf ("MemberTable[%d].RemoteSubscribedMsg=",Member_ID);
    
    for (int i = (NUMINT_MSGTYPE - 1); i >= 0; i--)
      printf ("[%x]", MemberTable[Member_ID].RemoteSubscribedMsg[i]);
    printf ("\n");
#endif
    
    // Update message subscription table
    offset = (Member_ID - 1) / SIZEOF_INTEGER;
    pos = Member_ID - (offset * SIZEOF_INTEGER);
    mask = ~(1 << (pos - 1));
    
    SubscribedTable[MsgType].mtx.lock ();
    
    // Local messages
    if ( ( myAgora->CheckMemberProfile(Member_ID, MPM_USER ) && ( ReqType & DCDT_LOCAL_MSG ) ) ||
	 ( myAgora->CheckMemberProfile(Member_ID, MPM_BRIDGE ) ) ||
	 ( myAgora->CheckMemberProfile(Member_ID, MPM_LINK ) ) )
      SubscribedTable[MsgType].SubscribersIfLocal[offset] &= mask;

    // Remote messages
    if ( ( myAgora->CheckMemberProfile(Member_ID, MPM_BRIDGE ) ) ||
	 ( ( ReqType & DCDT_REMOTE_MSG ) && (myAgora->CheckMemberProfile(Member_ID, MPM_USER) ) ) )
      SubscribedTable[MsgType].SubscribersIfRemote[offset] &= mask;

    SubscribedTable[MsgType].mtx.unlock ();

    // Update the update flags    
    UpdateMask_mtx.lock();
    for (int i=0; i<Member_ID; i++)
      MemberTable[i].update = true;
    for (int i=Member_ID+1; i<MAX_MEMBERS; i++)
      MemberTable[i].update = true;
    UpdateMask_flag = true;
    UpdateMask_mtx.unlock();
    
  }

  // TODO Check if there isn't messages with the same Member_ID to be deleted
};

/*! \brief Unsubscribe the member \a Member_ID from all the messages types
 *
 */
inline void DCDT_PostOffice_SLWBU::UnSubscribeAll ( int Member_ID )
{
  int i, j, offset, pos, MsgType;
  unsigned int msg_mask, mask;
  
  TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("UnSubscription request from Member %d to All Msg Types", Member_ID));

  offset = (Member_ID - 1) / SIZEOF_INTEGER;
  pos = Member_ID - (offset * SIZEOF_INTEGER);
  mask = ~(1 << (pos - 1));
  
  // For all the messages:
  // - If we are susbscribed unsubscribe and update the message subscription mask
  for (i = 0; i < NUMINT_MSGTYPE; i++) {
    for (j = 0; j < SIZEOF_INTEGER; j++) {
      msg_mask = 1;
      msg_mask <<= j;
      if ( ( msg_mask & MemberTable[Member_ID].LocalSubscribedMsg[i] ) ||
	   ( msg_mask & MemberTable[Member_ID].RemoteSubscribedMsg[i] ) )
	{
	  MsgType = SIZEOF_INTEGER * i + j + 1;
	  SubscribedTable[MsgType].mtx.lock ();
	  
	  // Unsubscribe Member_ID from local and remote msg types
	  SubscribedTable[MsgType].SubscribersIfLocal[offset] &= mask;
	  SubscribedTable[MsgType].SubscribersIfRemote[offset] &= mask;
	  
	  SubscribedTable[MsgType].mtx.unlock ();
	}
    }
    
    MemberTable[Member_ID].LocalSubscribedMsg[i] = 0;
    MemberTable[Member_ID].RemoteSubscribedMsg[i] = 0;
  }
  
  UpdateMask_mtx.lock();
  for (int i=0; i<Member_ID; i++)
    MemberTable[i].update = true;
  for (int i=Member_ID+1; i<MAX_MEMBERS; i++)
    MemberTable[i].update = true;
  UpdateMask_flag = true;
  UpdateMask_mtx.unlock();
  
}


/*! \brief Set the type of the message point by \a Elem to \a num_type.
 *
 *  This method is used only during the insert phase so no lock is required because there aren't concurrency problems.
 */
inline void DCDT_PostOffice_SLWBU::SetType (DCDT_MsgListElem_SLWBU * Elem, int num_type)
{
  int offset, pos;

  if (Elem && ((num_type > 0) && (num_type <= MAX_MSG_TYPE))) {
    offset = (num_type - 1) / SIZEOF_INTEGER;
    pos = num_type - (offset * SIZEOF_INTEGER);
    
    Elem->type_mask[offset] |= 1 << (pos -1);
    
  }
  else {
    TRC_PRINT ( DCDT_TRC_POSTOFFICE, TRC1, ("SetType ERROR"));
  }
}



/*! \brief Update the SubscribedMsgMask used by LinkTx when talking with another Agora.
 *
 *  Rebuild the global subscribers mask; the global mask is build upon the single message mask of each active member, don't include the mask of the member \a MemID; if \a MemID is a bridge all the members masks are used otherwise only the masks of MPM_USER and MPM_BRIDGE type members are used; so in a LinkTx member we can build the subscriber mask that needs to be passed to the other side of the link.
 */
inline void DCDT_PostOffice_SLWBU::UpdateSubscribersMask (int MemID)
{
  int i, j, isBridge;
  unsigned int temp[NUMINT_MSGTYPE];
  
  for (j = 0; j < NUMINT_MSGTYPE; j++)
    temp[j] = 0;
  
  isBridge=myAgora->CheckMemberProfile( MemID, MPM_BRIDGE );
  
  // Build a new temporary mask
  for (i = 0; i < MAX_MEMBERS; i++)
    {
      if (MemberTable[i].alive && (i != MemID)) {
	// MemID == MPM_BRIDGE -> request any message needed by local members
        // MemID != MPM_BRIDGE -> request messages only for MPM_USER and MPM_BRIDGE members ( not dynamic links, MPM_LINK )
	
	if( isBridge ) {
	  /* MPM_BRIDGE -> relay for local msg use Local Mask */
	  if ( myAgora->CheckMemberProfile(i, MPM_USER ) ||
	       myAgora->CheckMemberProfile(i, MPM_BRIDGE ) ) {
	    for (j = 0; j < NUMINT_MSGTYPE; j++)
	      temp[j] |= MemberTable[i].LocalSubscribedMsg[j];
	  }
	}
	else {
	  /* !MPM_BRIDGE -> use Remote Mask */
	  if ( myAgora->CheckMemberProfile(i, MPM_USER ) ||
               myAgora->CheckMemberProfile(i, MPM_BRIDGE ) ) {
            for (j = 0; j < NUMINT_MSGTYPE; j++)
              temp[j] |= MemberTable[i].RemoteSubscribedMsg[j];
	  }
	}
      }
    }
  
  subscribed_mtx.lock ();
  
  // Now update the mask
  for (j = 0; j < NUMINT_MSGTYPE; j++)
    SubscribedMsgMask[j] = temp[j];
  
  subscribed_mtx.unlock ();  
};

/*! \brief Returns the decimal value corresponding to the mask signature of the element pointed by \a Elem.
 */
inline int DCDT_PostOffice_SLWBU::GetTypeFromMask (DCDT_MsgListElem_SLWBU * Elem)
{
  int i, j, temp;
  
  for (i = (NUMINT_MSGTYPE - 1); i >= 0; i--)
    {
      if (Elem->type_mask[i])
    	{
  	  temp = Elem->type_mask[i];
  	  j = 0;
  	  while (temp)
	    {
	      temp = temp >> 1;
	      j++;
	    };
	  return (i * SIZEOF_INTEGER + j);
	};
    };
};


/*! \brief Check if SubscribedMsgMask is changed.
 *
 *  \return TRUE - SubscribedMsgMask was changed
 *  \return FALSE - SubscribedMsgMask wasn't changed  
 */
inline bool DCDT_PostOffice_SLWBU::ReadSubscribedMsgMask(unsigned int *mask, int MemID) 
{
  bool match=true;
  
  UpdateMask_mtx.lock();
  if(MemberTable[MemID].update) 
    {
      UpdateSubscribersMask(MemID);
      MemberTable[MemID].update = false;
    };
  UpdateMask_mtx.unlock();
  
  for (int i=0; i<NUMINT_MSGTYPE; i++)
    match = match && (mask[i] == SubscribedMsgMask[i]);
  
  for (int i=0; i<NUMINT_MSGTYPE; i++)
    mask[i] = SubscribedMsgMask[i];

  // The old mask matches ?
  if (!match)
    return true;
  else
    return false;
}

inline bool DCDT_PostOffice_SLWBU::UpdateSubs(int MemID) 
{
  return MemberTable[MemID].update;
}

#ifdef TEST_PART

inline void DCDT_PostOffice_SLWBU::InsertListTime (int MemID, int Type_Op)
{
  DCDT_TIME x = GetTime();
  TimeTesterElem *newElem;
  newElem = new TimeTesterElem;
  newElem->Data = x;
  newElem->Type_Op = Type_Op;
  newElem->prev = TestElem[MemID].prev;
  TestElem[MemID].prev = newElem; 
} 


inline void DCDT_PostOffice_SLWBU::VisualList()
{
  TimeTesterElem *newElem;
  
  for (int i = 0; i<MAX_MEMBERS;i++) {
    newElem = TestElem[i].prev;
    while (newElem) {
      cout<<"["<<i<<","<<newElem->Type_Op<<","<<newElem->Data-TestElem[i].Data<<"]";
      newElem=newElem->prev;
    }
  } 
}

inline void DCDT_PostOffice_SLWBU::SaveList(char* nomefile)  
{
  TimeTesterElem *newElem; 
  FILE *fout;
  time_t longtime;
  struct tm* time_experiment;
  
  char string[100] = "DATI_";

  strcat(string,nomefile); 
  
  fout = fopen (string,"wb");

  // Insert time and date of the operation
  time(&longtime);
  time_experiment = localtime (&longtime);

  putw(time_experiment->tm_hour,fout);
  putw(time_experiment->tm_min,fout); 
  putw(time_experiment->tm_mday,fout); 
  putw(time_experiment->tm_mon,fout);
  putw(1900 + time_experiment->tm_year,fout);

  // Software version and system parameters
  putw(VERSION,fout);
  putw(NUMMEMBERFORMAIN,fout);

  // Note that  NUMMEMBERFORMAIN = NUMREADER + NUMWRITER + NUMREADERWRITER 
  putw(NUMREADER,fout);      
  putw(NUMWRITER,fout);      
  putw(NUMREADERWRITER,fout); 
  
  // Number of cycles  
  putw(READCYCLE,fout); 
  putw(WRITECYCLE,fout);        
  putw(READWRITECYCLE,fout);   
  
  for (int i = 0; i<MAX_MEMBERS;i++) {
    
    newElem = TestElem[i].prev;
    
    while (newElem) {
      putw(newElem->Type_Op,fout);
      putw(newElem->Data-TestElem[i].Data,fout);
      
      newElem=newElem->prev;
    }
    putw(666,fout);
  }

  fclose(fout);
}

#endif /* TEST_PART */

#endif /* DCDT_POSTOFFICE_SLWBU_H */
