/*
 * AIRboardUI: graphical interface program to interact with airboard
 * through RS232 serial port.
 *
 * In order to compile this package you have to install the
 * Java Communications API on your system.
 * http://java.sun.com/products/javacomm/downloads/index.html
 * 
 * Bruschieri Stefano
 * Campari PierPaolo
 */

package airboardUI;
import java.awt.Choice;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.GridLayout;
import java.awt.Label;
import java.awt.Button;
import java.awt.Panel;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.util.Enumeration;

import javax.comm.CommPortIdentifier;

/**
 * @author Stefano
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class SerialParametersForm extends Frame implements ActionListener {
    
	private static final int WIDTH = 400;
	private static final int HEIGHT = 150;
    private SerialParameters parameters;
    private SerialConnection connection;
    private ConfigurationPanel configurationPanel;
    
    private Button okButton;
    private Button cancelButton;
    private Panel buttonPanel;
    private AirboardFrame parent;
    
    public boolean portChoosen = false;
	
	public SerialParametersForm(SerialParameters parameters, 
			SerialConnection connection,
			AirboardFrame parent){
		super("Settings");
		this.parameters = parameters;
		this.connection = connection;
		this.parent = parent;
		addWindowListener(new CloseHandler(this));
		
		configurationPanel = new ConfigurationPanel(this);
		add(configurationPanel);
		
		buttonPanel = new Panel();
		
		okButton = new Button("Apply");
		okButton.addActionListener(this);
		buttonPanel.add(okButton);
		cancelButton = new Button("Cancel");
		cancelButton.addActionListener(this);
		buttonPanel.add(cancelButton);
		
		add(buttonPanel, "South");

		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
		this.setLocation(screenSize.width/2 - WIDTH/2, 
			    screenSize.height/2 - HEIGHT/2);
		this.setSize(WIDTH, HEIGHT);
		this.setVisible(true);
		this.repaint();
	}
	
    /**
    Responds to the buttons.
    */
    public void actionPerformed(ActionEvent e) {
    	String cmd = e.getActionCommand();
    	
    	if(cmd.equals("Apply")){
    		//Check if port is open.
    	    if (connection.isOpen()) {
    		// If port is open do not allow port to change.
    	    	if (portChoosen) {
    		    // Alert user.
    		    AlertDialog ad = new AlertDialog(this , "Port Open!",
    					 	     "Port can not"+
    						     "be changed"+
    						     "while a port is open.");

    		    // Return configurationPanel to pre-choice settings.
    		    configurationPanel.setConfigurationPanel();
    		    return;
    		}
    		// Set the parameters from the choice panel.
    	    configurationPanel.setParameters();
    		try {
    		    // Attempt to change the settings on an open port.
    		    connection.setConnectionParameters();
    		} catch (SerialConnectionException ex) {
    		    // If setting can not be changed, alert user, return to
    		    // pre-choice settings.
    		    AlertDialog ad = new AlertDialog(this , 
    			"Unsupported Configuration!",
    			"Configuration Parameter unsupported,"+
    			"select new value."+
    			"Returning to previous configuration.");
    		    configurationPanel.setConfigurationPanel();
    		}
    	    } else {
    		// Since port is not open just set the parameter object.
    	    	configurationPanel.setParameters();
    	    }
    	    parent.updateSettingsPanel();
    	    this.setVisible(false);
    		//SerialParametersForm spf = new SerialParametersForm(this.parameters, this.connection);
    	}
    	
    	//Setting window closes without saving
    	if(cmd.equals("Cancel")){
    		this.setVisible(false);
    	}
    }
	    
    /**
    Handles closing down system. Allows application to be closed with window
    close box.
    */
    class CloseHandler extends WindowAdapter {

	SerialParametersForm spf;
	
	public CloseHandler(SerialParametersForm spf) {
	    this.spf = spf;
	}

	public void windowClosing(WindowEvent e) {
		spf.setVisible(false);
	}
    }
    
    /**
    GUI element that holds the user changable elements for connection
    configuration.
    */
    class ConfigurationPanel extends Panel implements ItemListener {

	private Frame parent;

	private Label portNameLabel;
	private Choice portChoice;

	private Label baudLabel;
	private Choice baudChoice;

	private Label flowControlInLabel;
	private Choice flowChoiceIn;

	private Label flowControlOutLabel;
	private Choice flowChoiceOut;

	private Label databitsLabel;
    private Choice databitsChoice;

	private Label stopbitsLabel;
    private Choice stopbitsChoice;

	private Label parityLabel;
    private Choice parityChoice;

	/**
	Creates and initilizes the configuration panel. The initial settings
	are from the parameters object.
	*/
	public ConfigurationPanel(Frame parent) {
	    this.parent = parent;

	    setLayout(new GridLayout(4, 4));

	    portNameLabel = new Label("Port Name:", Label.LEFT);
	    add(portNameLabel);

	    portChoice = new Choice();
	    portChoice.addItemListener(this);
	    add(portChoice);
	    listPortChoices();
	    portChoice.select(parameters.getPortName());
	    
	    baudLabel = new Label("Baud Rate:", Label.LEFT);
	    add(baudLabel);

	    baudChoice = new Choice();
	    baudChoice.addItem("300");
	    baudChoice.addItem("2400");
	    baudChoice.addItem("9600");
	    baudChoice.addItem("14400");
	    baudChoice.addItem("28800");
	    baudChoice.addItem("38400");
	    baudChoice.addItem("57600");
	    baudChoice.addItem("152000");
	    baudChoice.select(Integer.toString(parameters.getBaudRate()));
	    baudChoice.addItemListener(this);
	    add(baudChoice);

	    flowControlInLabel = new Label("Flow Control In:", Label.LEFT);
	    add(flowControlInLabel);

	    flowChoiceIn = new Choice();
	    flowChoiceIn.addItem("None");
	    flowChoiceIn.addItem("Xon/Xoff In");
	    flowChoiceIn.addItem("RTS/CTS In");
   	    flowChoiceIn.select(parameters.getFlowControlInString());
	    flowChoiceIn.addItemListener(this);
	    add(flowChoiceIn);

	    flowControlOutLabel = new Label("Flow Control Out:", Label.LEFT);
	    add(flowControlOutLabel);

	    flowChoiceOut = new Choice();
	    flowChoiceOut.addItem("None");
	    flowChoiceOut.addItem("Xon/Xoff Out");
	    flowChoiceOut.addItem("RTS/CTS Out");
	    flowChoiceOut.select(parameters.getFlowControlOutString());
	    flowChoiceOut.addItemListener(this);
	    add(flowChoiceOut);

	    databitsLabel = new Label("Data Bits:", Label.LEFT);
	    add(databitsLabel);

	    databitsChoice = new Choice();
	    databitsChoice.addItem("5");
	    databitsChoice.addItem("6");
	    databitsChoice.addItem("7");
	    databitsChoice.addItem("8");
	    databitsChoice.select(parameters.getDatabitsString());
	    databitsChoice.addItemListener(this);
	    add(databitsChoice);

	    stopbitsLabel = new Label("Stop Bits:", Label.LEFT);
	    add(stopbitsLabel);

	    stopbitsChoice = new Choice();
	    stopbitsChoice.addItem("1");
	    stopbitsChoice.addItem("1.5");
	    stopbitsChoice.addItem("2");
	    stopbitsChoice.select(parameters.getStopbitsString());
	    stopbitsChoice.addItemListener(this);
 	    add(stopbitsChoice);

	    parityLabel = new Label("Parity:", Label.LEFT);
	    add(parityLabel);
	    
	    parityChoice = new Choice();
	    parityChoice.addItem("None");
	    parityChoice.addItem("Even");
	    parityChoice.addItem("Odd");
	    parityChoice.select("None");
	    parityChoice.select(parameters.getParityString());
	    parityChoice.addItemListener(this);
	    add(parityChoice);
	}

	/**
	Sets the configuration panel to the settings in the parameters object.
	*/
	public void setConfigurationPanel() {
	    portChoice.select(parameters.getPortName());
	    baudChoice.select(parameters.getBaudRateString());
	    flowChoiceIn.select(parameters.getFlowControlInString());
	    flowChoiceOut.select(parameters.getFlowControlOutString());
	    databitsChoice.select(parameters.getDatabitsString());
	    stopbitsChoice.select(parameters.getStopbitsString());
	    parityChoice.select(parameters.getParityString());
	}

	/**
	Sets the parameters object to the settings in the configuration panel.
	*/
        public void setParameters() {
            parameters.setPortName(portChoice.getSelectedItem());
            //System.out.println(portChoice.getSelectedItem());
            parameters.setBaudRate(baudChoice.getSelectedItem());
            parameters.setFlowControlIn(flowChoiceIn.getSelectedItem());
            parameters.setFlowControlOut(flowChoiceOut.getSelectedItem());
            parameters.setDatabits(databitsChoice.getSelectedItem());
            parameters.setStopbits(stopbitsChoice.getSelectedItem());
            parameters.setParity(parityChoice.getSelectedItem());
        }

	/**
	Sets the elements for the portChoice from the ports available on the
	system. Uses an emuneration of comm ports returned by 
	CommPortIdentifier.getPortIdentifiers(), then sets the current
	choice to a mathing element in the parameters object.
	*/
        void listPortChoices() {
            CommPortIdentifier portId;

            Enumeration en = CommPortIdentifier.getPortIdentifiers();

            // iterate through the ports.
            while (en.hasMoreElements()) {
                portId = (CommPortIdentifier) en.nextElement();
                if (portId.getPortType() == CommPortIdentifier.PORT_SERIAL) {
                    portChoice.addItem(portId.getName());
                }
            }
            portChoice.select(parameters.getPortName());
        }

	/**
	Event handler for changes in the current selection of the Choices.
	If a port is open the port can not be changed.
	If the choice is unsupported on the platform then the user will
	be notified and the settings will revert to their pre-selection
	state.
	*/
	public void itemStateChanged(ItemEvent e) {
		
		if (e.getItemSelectable() == portChoice) {
			portChoosen = true;
		}
		
		/*
	    // Check if port is open.
	    if (connection.isOpen()) {
		// If port is open do not allow port to change.
	    	if (e.getItemSelectable() == portChoice) {
		    // Alert user.
		    AlertDialog ad = new AlertDialog(parent, "Port Open!",
					 	     "Port can not", 
						     "be changed",
						     "while a port is open.");

		    // Return configurationPanel to pre-choice settings.
		    setConfigurationPanel();
		    return;
		}
		// Set the parameters from the choice panel.
		setParameters();
		try {
		    // Attempt to change the settings on an open port.
		    connection.setConnectionParameters();
		} catch (SerialConnectionException ex) {
		    // If setting can not be changed, alert user, return to
		    // pre-choice settings.
		    AlertDialog ad = new AlertDialog(parent, 
			"Unsupported Configuration!",
			"Configuration Parameter unsupported,",
			"select new value.",
			"Returning to previous configuration.");
		    setConfigurationPanel();
		}
	    } else {
		// Since port is not open just set the parameter object.
	    	setParameters();
	    }
	    */
	}
    }

}
