#ifndef MetamorphicActuationManager_h
#define MetamorphicActuationManager_h

#include "ActuationManager.h"

/**
 * This class implements the translation of actions and odometry for robots
 * with the Metamorphic kinematics (i.e., two parallel wheels that are fixed 
 * to the chassis.
 *
 * The MetamorphicActuationManager class implements the method for translating the
 * actions received by the high-level controller and the method for computing
 * the robot movement starting from the odometric data acquired by the low-level
 * control board, according to the Giano kinematics. In particular, the
 * high-level controller will produce commands for the speed along the x and y 
 * axes and for the rotational speed that must be translated in commands for the
 * motors. On the other hand, from the readings of the encoders this class will 
 * produce the actual movement of the robot.
 *
 * @short	Implementation of the ActuationManager for the Giano kinematics
 * @author	Marcello Restelli (restelli@elet.polimi.it)
 */

class MetamorphicActuationManager : public ActuationManager
{
 public:
  // LIFECYCLE

  /**
   * Default constructor
   */
  MetamorphicActuationManager();

  /**
   * Destructor
   */
  virtual ~MetamorphicActuationManager();


  // OPERATORS

  //Use compiler-generated copy constructor and assignment
  //MetamorphicActuationManager(const MetamorphicActuationManager& from);
  //MetamorphicActuationManager& operator=(const MetamorphicActuationManager& from);

  // OPERATIONS

  /**
   * Decompose the actions received by the high-level controller
   * into commands for the low-level controller
   */ 
  void ComputeActuations();

  /**
   * Starting from the data of the encoders, this method computes the robot
   * movement
   */
  void ComputeOdometry();
};

#endif
