#ifndef SerialCommunicationAbaco_h
#define SerialCommunicationAbaco_h

#include "SerialCommunication.h"


/**
 * This class implements the communication thorugh serial device toward the
 * control board produced by Abaco.
 *
 * The SerialCommunicationAbaco class implements the methods to write to and
 * read from the Abaco control board. In particular, this class implements
 * two pure virtual methods defined in the @ref LowLevelCommunication interface.
 *
 * @short	Interface for the Abaco control board
 * @author	Marcello Restelli (restelli@elet.polimi.it)
 */

class SerialCommunicationAbaco : public SerialCommunication
{
 public:
  // LIFECYCLE

  /**
   * Default constructor
   */
  SerialCommunicationAbaco();

  /**
   * Destructor
   */
  virtual ~SerialCommunicationAbaco();


  // OPERATORS

  //Use compiler-generated copy constructor and assignment
  //SerialCommunicationAbaco(const SerialCommunicationAbaco& from);
  //SerialCommunicationAbaco& operator=(const SerialCommunicationAbaco& from);

  // OPERATIONS
  

  /**
   * This is a pure virtual method that communicates the data stored in 
   * mActuationValues to the Abaco's board
   */
  void WriteActuations();

  /**
   * This is a pure virtual method that receives the data sent by the Abaco's
   * board and stores them in mEncoderValues
   */
  void ReadEncoders();
};

#endif
