#ifndef SynchroActuationManager_h
#define SynchroActuationManager_h

#include "ActuationManager.h"

/**
 * This class implements the translation of actions and odometry for robots
 * with the Synchro kinematics (i.e., two parallel wheels that are fixed to the
 * chassis).
 *
 * The SynchroActuationManager class implements the method for translating the
 * actions received by the high-level controller and the method for computing
 * the robot movement starting from the odometric data acquired by the low-level
 * control board, according to the Synchro kinematics. In particular, the
 * high-level controller will produce commands for the tangential speed and for
 * the rotational speed that must be translated in commands for the motors. On
 * the other hand, from the readings of the encoders this class will produce the
 * actual movement of the robot.
 *
 * @short	Implementation of the ActuationManager for the Synchro kinematics
 * @author	Marcello Restelli (restelli@elet.polimi.it)
 */

class SynchroActuationManager : public ActuationManager
{
 public:
  // LIFECYCLE

  /**
   * Default constructor
   */
  SynchroActuationManager();

  /**
   * Destructor
   */
  virtual ~SynchroActuationManager();


  // OPERATORS

  //Use compiler-generated copy constructor and assignment
  //SynchroActuationManager(const SynchroActuationManager& from);
  //SynchroActuationManager& operator=(const SynchroActuationManager& from);

  // OPERATIONS

  /**
   * Decompose the actions received by the high-level controller
   * into commands for the low-level controller
   */ 
  void ComputeActuations();

  /**
   * Starting from the data of the encoders, this method computes the robot
   * movement
   */
  void ComputeOdometry();

 private:
  float mWheelDistance;
};

#endif
