#include "TriskarActuationManager.h"
#include "MotorConstants.h"

#include <math.h>

const float rad3 = 1.73205;
const float DEG2RAD = M_PI/180.0;

TriskarActuationManager::TriskarActuationManager()
{}

TriskarActuationManager::~TriskarActuationManager()
{}


void TriskarActuationManager::ComputeActuations()
{ 
  mActuationValues.clear();
    
  if (mCommandData.find(TAN_X) == mCommandData.end())
    {
      mCommandData[TAN_X] = 0.0;
    }

  if (mCommandData.find(TAN_Y) == mCommandData.end())
    {
      mCommandData[TAN_Y] = 0.0;
    }  
 
  if (mCommandData.find(ROT) == mCommandData.end())
    {
      mCommandData[ROT] = 0.0;
    }

  if (mCommandData.find(SPEED_MODULE) != mCommandData.end()
      &&
      mCommandData.find(SPEED_ANGLE) != mCommandData.end())
    {
      mCommandData[TAN_X] = mCommandData[SPEED_MODULE]*cosf(mCommandData[SPEED_ANGLE]*DEG2RAD);
      mCommandData[TAN_Y] = mCommandData[SPEED_MODULE]*sinf(mCommandData[SPEED_ANGLE]*DEG2RAD);
    }

  
  float speed_0 = (-mCommandData[TAN_X]   + mCommandData[TAN_Y] / rad3			 + mCommandData[ROT]);
  float speed_1 = (                         mCommandData[TAN_Y] * - 2/rad3			 + mCommandData[ROT]);
  float speed_2 = ( mCommandData[TAN_X]   + mCommandData[TAN_Y] / rad3			 + mCommandData[ROT]);



  float max_speed = max(fabs(speed_0),max(fabs(speed_1),fabs(speed_2)));
  if (max_speed < 100)
  {
    max_speed = 1.0;
  }
  else
  {
    max_speed /= 100.0;
  }


  mActuationValues[SPEED_MOTOR_0] = (int)(speed_0/max_speed);
  mActuationValues[SPEED_MOTOR_1] = (int)(speed_1/max_speed);
  mActuationValues[SPEED_MOTOR_2] = (int)(speed_2/max_speed);


  if (mCommandData.find(KICK) != mCommandData.end())
    {
      mActuationValues[KICK] = (int)mCommandData[KICK];
    }
    
  if (mCommandData.find(STATUS) != mCommandData.end())
    {
      mActuationValues[STATUS] = (int)mCommandData[STATUS];
    }
}

void TriskarActuationManager::ComputeOdometry()
{
  float x   = (1/3 * (mEncoderValues[ENCODER_MOTOR_0] * (-60 / rad3) 
		      + mEncoderValues[ENCODER_MOTOR_2] * (60 / rad3) ));
  float y   = (1/3 * (mEncoderValues[ENCODER_MOTOR_0] * 20
		      + mEncoderValues[ENCODER_MOTOR_1] * -40 
		      + mEncoderValues[ENCODER_MOTOR_2] *  20         ));
  float phi = (1/3 * (mEncoderValues[ENCODER_MOTOR_0] *  -1
		      + mEncoderValues[ENCODER_MOTOR_1] * -1
		      + mEncoderValues[ENCODER_MOTOR_2] *  -1         ));
  
  mOdometryData[ROBOT_X]   = x  ;
  mOdometryData[ROBOT_Y]   = y  ;
  mOdometryData[ROBOT_PHI] = phi;
  
  mOdometryData[ACT_TAN_X_SPEED] = (float)(1/3 * (mEncoderValues[SPEED_MOTOR_0] * (-60 / rad3)
						  + mEncoderValues[SPEED_MOTOR_2] * (60 / rad3) ));
  mOdometryData[ACT_TAN_Y_SPEED] = (float)(1/3 * (mEncoderValues[SPEED_MOTOR_0] *   20 
						  + mEncoderValues[SPEED_MOTOR_1] * -40
						  + mEncoderValues[SPEED_MOTOR_2] *  20));
  mOdometryData[ACT_ROT_SPEED]   = (float)(1/3 * (mEncoderValues[SPEED_MOTOR_0] *  -1 
						  + mEncoderValues[SPEED_MOTOR_1] * -1
						  + mEncoderValues[SPEED_MOTOR_2] *  -1         ));
}
