# -*- coding: UTF-8 -*-

import sys
import getopt
import re
import fancyMsgs

#contants

directions = ['N','NNW','NW','WNW','W','WSW','SW','SSW','S','SSE','SE','ESE','E','ENE','NE','NNE']

RESET = '\033[0;0m'
BOLD = '\033[1m'
BLACK = '\033[30m'
RED = '\033[31m'
GREEN = '\033[32m'
YELLOW = '\033[33m'
BLUE = '\033[34m'
MAGENTA = '\033[35m'
CYAN = '\033[36m'
WHITE = '\033[37m'

def usage():
	print 'This script generates rules for Mr.BRIAN engine starting from templates'
	print 'Version 0.2	'
	print '2006-05-32'
	print '\n'
	
	
	print '\t-h, --help\tPrint this message\n'
	
	print '\t-l, --log\tAdd comments to the output file to identify the auto generated rules\n'
	
	print '\t-o, --output\tThe output file generated from the template\n'
	
	print '\t-i, --input\tThe input file containing the template rules\n'
	print 'Example of template (TODO)'
	print '[4]{'
	print 'This is  <D>'
	print 'Angle+1 <D+1>'
	print 'Angle-1 <D-1>'
	print '}'
	
def getFile(fileName):
	try:
		f = open(fileName, "rb")
		text = f.read()
		f.close()
		return text
	except IOError:
		print '\t',RED,BOLD,'ERROR',RESET,' cannot open file ' + fileName
		sys.exit()

def getElement(array,offset):
	# fare il modulo dell'offset 
	if offset < 0:
		offset = offset + len(array) 
	return array[offset]

def shiftArray(array, n):
	for i in range(0,n):
		temp = array[0]
		for j in range(0, len(array)-1):
			array[j] = array[j+1]
		array[len(array)-1] = temp

def defineRegex(robot):
	if (robot=='giano'):
		#get simmetry family
		#S = standard simmetry
		simmetryRE = '\[(?P<simmetry>[\s|S]*)\]'
	
		#whitespaces
		sepRE = '[\s]*'
	
		#get first Template secrion
		tempRE1 ='\{(?P<template1>[\w\W]*?)\}'
		#get second Template section
		tempRE2 ='\{(?P<template2>[\w\W]*?)\}'
		
		return simmetryRE + sepRE + tempRE1 + sepRE + tempRE2 
	else:
		print 'ERROR'

#giano simmetries

def removeTags(text):
	for index in range(0,len(directions)):
		direction = directions[index]
		text = text.replace('<' + direction + '>',direction)
		
	text = text.replace('<VERY_FAST_','VERY_FAST_')
	text = text.replace('<FAST_','FAST_')
	text = text.replace('<RIGHT','RIGHT')
	text = text.replace('RIGHT>','RIGHT')
	text = text.replace('<LEFT','LEFT')
	text = text.replace('LEFT>','LEFT')
	text = text.replace('<SLOW_','SLOW_')
	text = text.replace('<VERY_SLOW_','VERY_SLOW_')
	
	text = text.replace('<FORWARD','FORWARD')
	text = text.replace('FORWARD>','FORWARD')
	text = text.replace('<BACKWARD','BACKWARD')
	text = text.replace('BACKWARD>','BACKWARD')
	
	text = text.replace('<TanFwd>','TanFwd')
	text = text.replace('<TanBwd>','TanBwd')
	
	text = text.replace('<RotRx>','RotRx')
	text = text.replace('<RotLx>','RotLx')
	
	return text
	
def mirrorAnglesWrtXAxis(text):
	for index in range(0,len(directions)):
		direction = directions[index]
		newIndex = (4*2 - index) % 16
		newDirection = directions[newIndex]
		text = text.replace('<' + direction + '>',newDirection)
			
	return text
	
def mirrorAnglesWrtYAxis(text):
	for index in range(0,len(directions)):
		direction = directions[index]
		newIndex = (16 - index) % 16
		newDirection = directions[newIndex]
		text = text.replace('<' + direction + '>',newDirection)	
	return text

def mirrorAnglesWrtCenter(text):
	for index in range(0,len(directions)):
		direction = directions[index]
		newIndex = (index + 8) % 16
		newDirection = directions[newIndex]
		text = text.replace('<' + direction + '>',newDirection)
	return text

def invertRotations(text):
	text = text.replace('<VERY_FAST_RIGHT>','VERY_FAST_LEFT')
	text = text.replace('<FAST_RIGHT>','FAST_LEFT')
	text = text.replace('<RIGHT>','LEFT')
	text = text.replace('<SLOW_RIGHT>','SLOW_LEFT')
	text = text.replace('<VERY_SLOW_RIGHT>','VERY_SLOW_LEFT')
	
	text = text.replace('<VERY_SLOW_LEFT>','VERY_SLOW_RIGHT')
	text = text.replace('<SLOW_LEFT>','SLOW_RIGHT')
	text = text.replace('<LEFT>','RIGHT')
	text = text.replace('<FAST_LEFT>','FAST_RIGHT')
	text = text.replace('<VERY_FAST_LEFT>','VERY_FAST_RIGHT')
	
	text = text.replace('<RotRx>','RotLx')
	text = text.replace('<RotLx>','RotRx')
	
	return text

def mirrorDirections(text):
	text = text.replace('<VERY_FAST_FORWARD>','VERY_FAST_BACKWARD')
	text = text.replace('<FAST_FORWARD>','FAST_BACKWARD')
	text = text.replace('<FORWARD>','BACKWARD')
	text = text.replace('<SLOW_FORWARD>','SLOW_BACKWARD')
	text = text.replace('<VERY_SLOW_FORWARD>','VERY_SLOW_BACKWARD')
	
	text = text.replace('<VERY_SLOW_BACKWARD>','VERY_SLOW_FORWARD')
	text = text.replace('<SLOW_BACKWARD>','SLOW_FORWARD')
	text = text.replace('<BACKWARD>','FORWARD')
	text = text.replace('<FAST_BACKWARD>','FAST_FORWARD')
	text = text.replace('<VERY_FAST_BACKWARD>','VERY_FAST_FORWARD')
	
	text = text.replace('<TanFwd>','TanBwd')
	text = text.replace('<TanBwd>','TanFwd')
	
	return text	
	
		
def buildGianoBehavior(inputFile, outputFile, log):
	print 'Generating from template file',BOLD,inputFile, RESET,'output file',BOLD,outputFile,RESET
	
	try:
		f = open(inputFile)
		f.close()			
	except IOError:
		print '\t',fancyMsgs.errorMsg('Cannot open input file ' + inputFile)
		#usage()
		sys.exit()
		
	if outputFile=='':
		print '\t',fancyMsgs.errorMsg('Please select output file')
		#usage()
		sys.exit()
		
	print '\tBuilding behavior...'
	gianoBehavior = ''
	
	text = getFile(inputFile)
	regex = defineRegex('giano')

	p = re.compile(regex,re.IGNORECASE)
	iterator = p.finditer(text)
	
	start = 0
	#count number of processed rules
	num = 1
	for match in iterator:
		print '\tGenerating section' , num ,
		#print '--Match found: ' , match.group()
		#print '--Match span: ' , match.span()
		#print '--Match group directions: ' , match.group('directions')
		simmetry = match.group('simmetry')
		#print '--Match group body: ' , match.group('template')
		template1 = match.group('template1')
		template2 = match.group('template2')
		
		gianoBehavior += text[start:match.start()]
		start = match.end()
				
		if log=='true':
			gianoBehavior += '## Start auto generated rule\n'
		if len(simmetry.strip()) > 0:
			num = num + 1
			print '\t','Simmetry type:' , simmetry 
		
		rule = ''
				
		#simmetry S:
		rule = rule + removeTags(template1)
		rule = rule + removeTags(template2)
		
		#template1 simmetric wrt the center
		behavior = template1
		behavior = mirrorDirections(behavior);
		behavior = mirrorAnglesWrtCenter(behavior);
		rule = rule + removeTags(behavior)
		
		#template2 simmetric wrt the y axis
		behavior = template2
		behavior = invertRotations(behavior);
		behavior = mirrorAnglesWrtYAxis(behavior);
		rule = rule + removeTags(behavior)
		
		#template2 simmetric wrt the x axis
		behavior = template2
		behavior = invertRotations(behavior);
		behavior = mirrorDirections(behavior);
		behavior = mirrorAnglesWrtXAxis(behavior);
		rule = rule + removeTags(behavior)
		
		#template2 simmetric wrt the center
		behavior = template2		
		behavior = mirrorAnglesWrtCenter(behavior);
		behavior = mirrorDirections(behavior);
		rule = rule + removeTags(behavior)
		
		if log=='true':
			gianoBehavior += '\n'
			gianoBehavior += '## Auto generated rule: simmetry type ' + simmetry + '\n'
			#gianoBehavior += '\n'
		gianoBehavior += '\n'
		gianoBehavior += rule
		if log=='true':
			gianoBehavior += '## End auto generated rule\n'
	
	gianoBehavior += text[start:]
	
	try:
		flag = 'false'
		try:
			f = open(outputFile, "rb")
			oldResult = f.read()
			f.close()
		except IOError:
			flag = 'true'
			oldResult = ''
			
		if gianoBehavior!=oldResult:
			if flag=='false':
				response = raw_input('\tOverride output file ' + outputFile + '? y/n ')
			else:
				response = 'y'
			if (response.strip().lower()=='y'):
				if flag=='false':
					print '\t',fancyMsgs.logMsg('File ' + outputFile + ' will be replaced')
				file = open(outputFile, "w")
				file.write(gianoBehavior)				
				print '\t',fancyMsgs.logMsg('File ' + outputFile + ' has been created successfully')
				file.close()
			else:
				print '\t',fancyMsgs.warningMsg('File will NOT be replaced')
				
			
		else:
			print '\t',fancyMsgs.logMsg('File ' + outputFile + ' is up to date')
	except IOError:
		print '\t',fancyMsgs.errorMsg('Cannot override output file' + outputFile)
		sys.exit()
		
def buildTriskarBehavior(inputFile, outputFile, log):
	allDirections = ['N','NNW','NW','WNW','W','WSW','SW','SSW','S','SSE','SE','ESE','E','ENE','NE','NNE']
	originalDirections = ['N','NNW','NW','WNW','W','WSW','SW','SSW','S','SSE','SE','ESE','E','ENE','NE','NNE']

	print 'Generating from template file',BOLD,inputFile, RESET,'output file',BOLD,outputFile,RESET
	
	try:
		f = open(inputFile)
		f.close()			
	except IOError:
		print '\t',RED,BOLD,'ERROR',RESET,' cannot open input file ' , inputFile , '\n'
		#usage()
		sys.exit()
		
	if outputFile=='':
		print '\t',RED,BOLD,'ERROR',RESET,' please select output file ' , '\n'
		#usage()
		sys.exit()
		
	print '\tBuilding behavior...'
	result = ''
	
	text = getFile(inputFile)
	#print text
	#print "Matching.."
	
	dirRE = '\[(?P<directions>[NSEW\-,\s48(16)]*)\]'
	sepRE = '[\s]*'
	tempRE ='\{(?P<template>[\w\W]*?)\}'
	p = re.compile(dirRE + sepRE + tempRE,re.IGNORECASE)
	iterator = p.finditer(text)
	
	start = 0
	num = 1
	for match in iterator:
		print '\tGenerating section ' , num ,
		#print '--Match found: ' , match.group()
		#print '--Match span: ' , match.span()
		#print '--Match group directions: ' , match.group('directions')
		directions = match.group('directions')
		#print '--Match group body: ' , match.group('template')
		template = match.group('template')
		
		result += text[start:match.start()]
		start = match.end()
		
		directionsToBuild = []
		
		r = re.compile(',',re.IGNORECASE)
		strings = r.split(directions)
		for string in strings:
			#print '--------Match found: \"' , string , '\"'
			if string.strip()=='4' or string.strip()=='8':
				if 'N' not in directionsToBuild:
					directionsToBuild.append('N')
				if 'S' not in directionsToBuild:
					directionsToBuild.append('S')
				if 'E' not in directionsToBuild:
					directionsToBuild.append('E')
				if 'W' not in directionsToBuild:
					directionsToBuild.append('W')
			if string.strip()=='8':
				if 'NW' not in directionsToBuild:
					directionsToBuild.append('NW')
				if 'SW' not in directionsToBuild:
					directionsToBuild.append('SW')
				if 'SE' not in directionsToBuild:
					directionsToBuild.append('SE')
				if 'NE' not in directionsToBuild:
					directionsToBuild.append('NE')
			if string.strip()=='16':
				directionsToBuild = allDirections
			else:
				q = re.compile('\-',re.IGNORECASE)
				iStrings = q.split(string)
				
				if len(iStrings)==1:
					#print '----------------------------Parsing: \"' , iStrings[0] , '\"'
					z = re.compile('[NSEW]+',re.IGNORECASE)
					iDirectionMatch = z.search(iStrings[0])
					
					if iDirectionMatch:
						iDirection = iDirectionMatch.group().strip()
						if iDirection not in allDirections:
							print '\t',RED,BOLD,'ERROR',RESET,' invalid direction \"' , iDirection , '\"'
							sys.exit()
						
						#print '----------------------------Match found: \"' , iDirection , '\"'
						if iDirection not in directionsToBuild:
							#print '----------------------------New direction: \"' , iDirection , '\"'
							directionsToBuild.append(iDirection)
				
				if len(iStrings)==2:
					#print '----------------------------Parsing: \"' , iStrings[0] , "-" , iStrings[1] , '\"'
					z = re.compile('[NSEW]+',re.IGNORECASE)
					
					startingDirectionMatch = z.search(iStrings[0])
					endingDirectionMatch = z.search(iStrings[1])
					
					if startingDirectionMatch:
						sDirection = startingDirectionMatch.group().strip()
						sIndex = allDirections.index(sDirection)
						if sDirection not in allDirections:
							print '\t',RED,BOLD,'ERROR',RESET,' invalid direction \"' , sDirection , '\"'
							sys.exit()
					
						if endingDirectionMatch:
							eDirection = endingDirectionMatch.group().strip()
							eIndex = allDirections.index(eDirection)
							if eDirection not in allDirections:
								print '\t',RED,BOLD,'ERROR',RESET,' invalid direction \"' , eDirection , '\"'
								sys.exit()
					
					
							if sIndex > eIndex:
								print '\t',RED,BOLD,'ERROR',RESET,' invalid direction interval: '
								sys.exit()
							else:
								for i in range(sIndex,eIndex):
									iDirection = allDirections[i]
									#print '----------------------------Match found: \"' , iDirection , '\"'
									if iDirection not in directionsToBuild:
										#print '----------------------------New direction: \"' , iDirection , '\"'
										directionsToBuild.append(iDirection)
		
		#print directionsToBuild
		for i in range(0,len(directionsToBuild)):
			for j in range(i+1,len(directionsToBuild)):
				if originalDirections.index(directionsToBuild[i]) > originalDirections.index(directionsToBuild[j]):
					#print 'prima' ,directionsToBuild[i] , directionsToBuild[j]
					temp = directionsToBuild[i]
					directionsToBuild[i] = directionsToBuild[j]
					directionsToBuild[j] = temp
					#print 'dopo' , directionsToBuild[i] , directionsToBuild[j]
		#print directionsToBuild
					
			
		#print directionsToBuild	
		if log=='true':
			result += '## Start auto generated rule\n'
		#print directionsToBuild ,
		if len(directionsToBuild) > 0:
			num = num + 1
		print directionsToBuild,
		#for direction in directionsToBuild:
		for index in range(0,len(directionsToBuild)):
			#print 'Indice' , index, 'Valore', directionsToBuild[index]
			direction = directionsToBuild[index]			 
			while direction != allDirections[0]:
				shiftArray(allDirections,1)
				#print "shifto, ora è " , allDirections
			# mettere le re per la sostituzione
			behavior = template.replace('<D>',allDirections[0])
			#print allDirections[0]
			#print originalDirections.index(allDirections[0])
			behavior = behavior.replace('<N>',str(originalDirections.index(allDirections[0])))
			for n in range(-15,16):
				# mettere le re per la sostituzione
				behavior = behavior.replace('<D' + str(n) + '>' , getElement(allDirections,n))
				if n>0:
					m = '+' + str(n) 
					# mettere le re per la sostituzione
					behavior = behavior.replace('<D' + str(m) + '>' , getElement(allDirections,n))
			if log=='true':
				result += '\n'
				result += '## Auto generated rule: direction ' + direction + '\n'
			result += behavior
			#result += '\n'
		result += '\n'
		print ''
		if log=='true':
			result += '## End auto generated rule\n'

	result += text[start:]
	
	try:
		flag = 'false'
		try:
			f = open(outputFile, "rb")
			oldResult = f.read()
			f.close()
		except IOError:
			flag = 'true'
			oldResult = ''
			
		if result!=oldResult:
			if flag=='false':
				response = raw_input('\tOverride output file ' + outputFile + '? y/n ')
			else:
				response = 'y'
			if (response.strip().lower()=='y'):
				if flag=='false':
					print '\t',YELLOW,BOLD,'WARNING',RESET,': the file ', outputFile,' will be replaced'
				file = open(outputFile, "w")
				file.write(result)				
				print '\t',GREEN,BOLD,'DONE',RESET,' the file',outputFile,'has been created successfully' ,				
				file.close()
			else:
				print '\t',RED,BOLD,'ERROR',RESET,' the file will NOT be replaced', '\n' ,	
			
		else:
			print '\t',GREEN,BOLD,'DONE',RESET,' the file',outputFile,'is up to date'
	except IOError:
		print '\t',RED,BOLD,'ERROR',RESET,' cannot write to output file ', outputFile ,
		sys.exit()


def main():
	inputFile = ''
	outputFile = ''
	log = ''
	robot  = ''
	
	try:
		options, arguments = getopt.getopt(sys.argv[1:], 'li:o:ht:', ['log','inputFile=','outputFile=','help','robot='])
	except getopt.GetoptError:
		print '\t',fancyMsgs.errorMsg('Error parsing arguments')
		sys.exit()
	log = 'false'
	for opt, arg in options:
		if opt in ("-l", "--log"):
			log = 'true'
		elif opt in ("-i", "--inputFile"):
			inputFile = arg
		elif opt in ("-o", "--outputFile"):
			outputFile = arg
		elif opt in ("-r", "--robot"):
			robot = arg
		elif opt in ("-h", "--help"):
			usage()
			sys.exit()
	
	if (robot=='giano'):
		buildGianoBehavior(inputFile,outputFile,log)
	elif (robot=='triskar'): 
		buildTriskarBehavior(inputFile,outputFile,log)
	else:
		print fancyMsgs.errorMsg('Missing argument. Please select robot type giano or triskar')

if __name__ == '__main__': main()
