#include "BrianExpert.h"
#include <iostream>
#include <stc.h>
#include <msg.h>
#include <MrtMisc.h>

#include "smdebug.h"

extern void brianlex_memory(const char* src, BrianExpert* brian);

//#define BRIAN_DEBUG 1 //*wiim*
//#define EXPERTS_DEBUG

#define FIELD_LENGTH 1200
#define AREA_X_LIMIT (FIELD_LENGTH/2) - 150 + 20
#define AREA_WIDENESS 500
#define AREA_Y_LIMIT (AREA_WIDENESS/2) + 20

BrianExpert::BrianExpert(ModuleDispatch* kernel,long period, bool refbox) : 
  StringModuleMember(kernel,period,"brian")
{
  if (!refbox)
    {
      mStatus = AUTO;
    }
  else
    {
      mStatus = MANUAL;
    }
  not_so_reliable = 0.9;
//  FILE * fd;

#ifdef EXPERTS_DEBUG
  cout<<"\n---------------> Starting Brian";
  fflush(stdout);
#endif

  TheBrian = new MrBrian(
  		"config/brian/2.1/ctof.txt",
  	 	"config/brian/2.1/shape_ctof.txt",
  	 	"config/brian/2.1/Predicate.ini",
  	 	"config/brian/2.1/PredicateActions.ini",
  	 	"config/brian/2.1/Cando.ini",
  	 	"config/brian/2.1/behaviour.txt",
  	 	"config/brian/2.1/want.txt",
  	 	"config/brian/2.1/s_ftoc.txt",
  	 	"config/brian/2.1/s_shapeFast.txt"
  	 	);
  	 	
  cout<<"\nMrBrian"<<endl;

  mRefbox = 0.0;
  old_ball_x = 10000;
  old_ball_y = 10000;
  mIsDefending = false;
  mIsWaiting = false;
  mWaitingTime = 0;
  mAreaTime = getTime();
}

BrianExpert::~BrianExpert()
{
  logfile.close();
  delete TheBrian;
}


void BrianExpert::RunInit()
{
  //AddMessageRequest(MSG_FROM_REFBOX,REMOTE);
  //AddMessageRequest(MSG_FROM_SCARE,LOCAL);
  //AddMessageRequest(MSG_FROM_WIIMOTE,LOCAL);
  // AddMessageRequest(MSG_FROM_MOTION,LOCAL);
  // AddMessageRequest(MSG_FROM_MAP,LOCAL);
  AddMessageRequest(MSG_TO_BRIAN,LOCAL);
  mAreaTime = getTime();
  
  logfile.open(BRIANFULLLOG);
}

void BrianExpert::RunClose()
{
}

void BrianExpert::RunDuty()
{

#ifdef SIM
  waitForTurn();
#endif
#ifdef EXPERTS_DEBUG    
  cout <<"\nBrian Starts";fflush(stdout); 
#endif  


  ParseStringMessages();

  TheBrian->run();
  
  //TheBrian->debug(); //*wiim*
  TheBrian->debugfile(&logfile);
  TheBrian->log();
  
  command_list * cl = (TheBrian->getFuzzy())->get_command_singleton_list();

  if (cl != 0)
    {
      command_list::iterator it;
      mCL.clear();
      
      for (it = cl->begin(); it != cl->end(); it++)
	{
	  mCL[it->first] = it->second->get_set_point();
	}
    }
    


  //  FillMessages(cl, linkacl, linkcadl, linkwwl, linkpdl);
  FillMessages((TheBrian->getFuzzy())->get_command_singleton_list(),
	       (TheBrian->getFuzzy())->get_action_list(),
               TheBrian->get_cando_list(),
               TheBrian->get_weight_want_list(),
               TheBrian->get_predicate_list());
 

 
  SendAllStringMessages();
  /* 
  viewcdl(cdl);

  viewcdl(cl);
  */
#ifdef BRIAN_DEBUG
  cout<<"\nDEBUG\n\n";
  //TheBrian->debug();
  printf("\nTIME: %d\n", TheBrian->get_time());
#endif

  TheBrian->flush();
  
#ifdef EXPERTS_DEBUG
  cout << "\n Brian ends";fflush(stdout);
#endif

#ifdef SIM
  endTurn();  
#endif

}

void BrianExpert::ParseStringMessages()
{

  unsigned char* buffer;

  if (ReceiveLastMessage(MSG_TO_BRIAN,buffer,true)>0){ 

FILE* fp=fopen(BRIANMSGLOG,"a");
fprintf(fp,"\n%s\n",buffer);
fclose(fp);
 	//printf ("ricevuto comando diretto a brian: \n%s\n",buffer);
    brianlex_memory((const char*)buffer,this);  
    
  }
/*
  if (ReceiveLastMessage(MSG_FROM_MAP,buffer,false)>0)
    brianlex_memory((const char*)buffer,this);
*/
/*  crisp_data_list::iterator ball_distance_it;
  ball_distance_it = cdl->find("BallDistance");
  float vision_ball_distance = 0.0;
  if (ball_distance_it != cdl->end())
    vision_ball_distance = ball_distance_it->second->get_value();
*/

/*  if (vision_ball_distance < 100)
    {
      SetValue("BallDistance",vision_ball_distance,1.0);
    }*/
/*
  if (ReceiveLastMessage(MSG_FROM_REFBOX,buffer,false)>0)
    brianlex_memory((const char*)buffer,this);

  if (ReceiveLastMessage(MSG_FROM_SCARE,buffer,false)>0)
    brianlex_memory((const char*)buffer,this);
*/
  // This is used to fix timing in waiting experts 
  ResetTimer();
}

void BrianExpert::FillMessages(command_list * cl, action_list * al, predicate_list* linkcadl, 
			       weight_want_list* linkwwl, predicate_list* linkpdl)
{
  // create the message to return
  NewMessage(MSG_TO_MOTION,COMMAND_DATA);
  
  command_multimap::iterator c_it;
  command* c = 0;
  for (c_it = cl->begin(); c_it != cl->end(); c_it++)
    {
      c = c_it->second;
      AddCommandData(c_it->first,c->get_set_point());

    }


//   NewMessage(MSG_FROM_BRIAN,BRIAN_DATA);
//   predicate_list::iterator i;
//   weight_want_list::iterator w;
//   for (i = linkcadl->begin(); i != linkcadl->end(); i++)
//     {
//       w = linkwwl->find(i->first);
//       AddBehaviorData(i->first,i->second->get_value(), w->second->get_value());
//     }

//    if (linkpdl != 0)
//     {
//       predicate_list::iterator p;

//       for (p = linkpdl->begin(); p != linkpdl->end(); p++)
// 	{
// 	  AddFuzzyData(p->first,p->second->get_value(),1.0);
// 	}
//     }
}


void BrianExpert::BuildData(command_list* cl)
{
  crisp_data_list::iterator ball_angle_it, ball_distance_it, home_it, goal_it,
    robot_x_it, robot_y_it, target_angle_it, refbox_it,
    ball_x_it, ball_y_it;
  
  crisp_data_list * cdl = (TheBrian->getFuzzy())->get_crisp_data_list();

  map<string,float>::iterator cl_it;
  for (cl_it = mCL.begin(); cl_it != mCL.end(); cl_it++)
    {
      SetValue(cl_it->first.data(),cl_it->second,1.0);
    }

  target_angle_it = cdl->find("OrientationTargetAngle");
  ball_angle_it = cdl->find("BallAngle");
  goal_it = cdl->find("GoalAngle");
  home_it = cdl->find("HomeAngle");
  refbox_it = cdl->find("RefereeBox");
  ball_x_it = cdl->find("BallX");
  ball_y_it = cdl->find("BallY");
  robot_x_it = cdl->find("RobotX");
  robot_y_it = cdl->find("RobotY");
  
  float ball_x = 0.0;
  if (ball_x_it != cdl->end())
    ball_x = ball_x_it->second->get_value();
  float ball_y  = 0.0;
  if (ball_y_it != cdl->end())
    ball_y = ball_y_it->second->get_value();
  float robot_x = 0.0;
  if (robot_x_it != cdl->end())
    robot_x = robot_x_it->second->get_value();
  float robot_y = 0.0;
  if (robot_y_it != cdl->end())
    robot_y  = robot_y_it->second->get_value();

  if ((robot_y > -AREA_Y_LIMIT) && (robot_y < AREA_Y_LIMIT) 
      &&
      ((robot_x > AREA_X_LIMIT) || (robot_x < -AREA_X_LIMIT)))
    {
      float time = (float)(getTime() - mAreaTime)/(float)1e6;
      SetValue("AreaTime",time, 1.0);
      cout << endl << "AreaTime " << time << endl;
    }
  else
    {
      mAreaTime = getTime();
    }

  float refboxtmp;

  if (refbox_it != cdl->end())
    {
      refboxtmp = refbox_it->second->get_value();
      if (refboxtmp < 0) // Stopping condition
	{
	  mStatus = 0;
	}
      else
	{
	  mStatus = AUTO;
	}


      if (refboxtmp > 0 && refboxtmp < 7) // The robot must defend
	{
	  mIsDefending = true;
	}

      if (refboxtmp == 0 && mIsDefending) // The game has started and we are defending
	{
	  if (ball_x_it != cdl->end())
	    {
	      old_ball_x = ball_x;
	    }
	  if (ball_y_it != cdl->end())
	    {
	      old_ball_y = ball_y;
	    }
	  mWaitingTime = getTime();
	  mIsDefending = false;
	  mIsWaiting = true;
	}
      else
	{
	  mRefbox = refboxtmp;
	}
    }
  //  cout << "RefBox " << getTime() - mWaitingTime << " " << sqrt((ball_x - old_ball_x)*(ball_x - old_ball_x) + (ball_y - old_ball_y)*(ball_y - old_ball_y)) << " " << waiting << endl;
  if (mIsWaiting && ((getTime() - mWaitingTime > 1e7)
		     || 
		     (sqrt((ball_x - old_ball_x)*(ball_x - old_ball_x) + (ball_y - old_ball_y)*(ball_y - old_ball_y))) > 50))
    {
      mIsWaiting = false;
      mRefbox = 0;
    }


  SetValue("Status",mStatus,1.0);
  SetValue("RefereeBox",mRefbox,1.0);
  SetValue("Role",0.0,1.0);

  if (ball_angle_it != cdl->end())
    {
      if (goal_it != cdl->end())
	{
	  float ball_goal_angle = goal_it->second->get_value() - ball_angle_it->second->get_value();
	  if (ball_goal_angle < 0)
	    {
	      ball_goal_angle += 360;
	    }
	  SetValue("BallGoalAngle",
		   ball_goal_angle,
		   ball_angle_it->second->get_reliability() * goal_it->second->get_reliability());
	}

      if (home_it != cdl->end())
	{
	  float ball_home_angle = home_it->second->get_value() - ball_angle_it->second->get_value();
	  if (ball_home_angle < 0)
	    {
	      ball_home_angle += 360;
	    }
	  SetValue("BallHomeAngle",
		   ball_home_angle,
		   ball_angle_it->second->get_reliability() * home_it->second->get_reliability());
	}
      if (target_angle_it != cdl->end())
	{
	  float ball_target_angle = target_angle_it->second->get_value() - ball_angle_it->second->get_value();
	  if (ball_target_angle < 0)
	    {
	      ball_target_angle += 360;
	    }
	  SetValue("BallTargetAngle",
		   ball_target_angle,
		   ball_angle_it->second->get_reliability() * target_angle_it->second->get_reliability());
	}
    }
  else
    {
      SetValue("BallDistance",-1.0,1.0);
      SetValue("BallAngle",-1.0,1.0);
    }

  
}

void BrianExpert::SetValue(const char * name, float val, float rel)
{
  // search for the crisp_data whose name is "name" and sets its value and reliability to "val" and "rel"

  crisp_data_list * cdl = (TheBrian->getFuzzy())->get_crisp_data_list();
  
  crisp_data_list::iterator i=cdl->find(name);
  if (i!=cdl->end())
    {
      (*i).second->set_value(val);
      (*i).second->set_reliability(rel);
    }
  else
    {
      cdl->add(new crisp_data(name,val,rel));
    }
}


void BrianExpert::AddBehaviorData(string name, 
					  float cando_value, 
					  float want_value)
{
  mCurrentMessage << "<B>"
		  << name << " " << cando_value << " " << want_value
		  << "</B>\n";
}

void BrianExpert::AddFuzzyData(string name, 
				       float value, 
				       float reliability)
{
  mCurrentMessage << "<F>"
		  << name << " " << value << " " << reliability
		  << "</F>\n";
}

void BrianExpert::AddCommandData(string name, float value)
{
	if (fabs(value)<0.001){//*wiim* evito che vengano mandati valori in notazione scientifica
		value=0;
	}
	
  mCurrentMessage << "<C>"
		  << name << " " << value
		  << "</C>\n";  
}
