#include "MotorExpert.h"
#include "MotorConstants.h"
#include "msg.h"

#include <iostream>

int needodo=0;

// la parola extern non dichiara un nuovo identificatore, 
// ma dice "La variabile intera a e` dichiarata da qualche altra parte, 
// lascia solo lo spazio per risolvere il riferimento"

extern void motorlex_memory(const char* src, MotorExpert* pMotor);
extern int MessageLineNum;

MotorExpert::MotorExpert(ModuleDispatch* kernel,int period, Spykee *s) : 
  StringModuleMember(kernel,period,"motor")
{
  spykee = s;
  printf("\nMotorExpert built");
  fflush(stdout);
};

MotorExpert::~MotorExpert()
{
  printf("\nMotorExpert destroyed");
  fflush(stdout);
};

void MotorExpert::RunInit()
{
	settimer(0);
	settimer(1);
	settimer(2);
	settimer(3);
	
  AddMessageRequest( MSG_TO_MOTION, LOCAL);
  
  SetWheelDistance(350);
  SetConversionCoefficients(0.0333,-0.0333);
  
  printf("\nMotorExpert initialized");
  fflush(stdout);
};

/* ***********************************
   
   DoYourDuty()

   *********************************** */

void MotorExpert::RunDuty()
{
	float Tan,Rot;
	int flag=0;

	//printf("\nSono motor");
	//fflush(stdout);

	// ************************************************************
	//  Se c'e' un messaggio da parte del controllo di alto livello, 
	//  cambio i setpoint, altrimenti cerco di mantenerli
	// ************************************************************

	unsigned char* buffer;
	if(ReceiveLastMessage(MSG_TO_MOTION, buffer,false) > 0)
	{
		struct timeval now;
		//printf("%d\n",elapsed(2));
		if(elapsed(1)<200000){
			return;
		}
		else
		{
			gettimeofday(&now, NULL);
			//printf("*****************nuovo comando: %d *      %d %d\n",elapsed(1),now.tv_sec,now.tv_usec);
			settimer(1);
		}
		//printf("*\n");
				
		mCommandData.clear();
		motorlex_memory((const char *)buffer,this);
		int dxSpeed = 1*((int)(mCommandData[TAN_SPEED] - mCommandData[ROT_SPEED]));
		int sxSpeed = 1*((int)(mCommandData[TAN_SPEED] + mCommandData[ROT_SPEED]));
		//ComputeActuations();
		//WriteActuations();
		if (dxSpeed > 90)
			dxSpeed = 90;
		if (sxSpeed > 90)
			sxSpeed = 90;
		if (dxSpeed < -90)
			dxSpeed = -90;
		if (sxSpeed < -90)
			sxSpeed = -90;
		spykee->move(dxSpeed, sxSpeed);
		NewMessage(MSG_FROM_MOTION,MOTION_DATA);
		AddOdometryData("TANSPEED", (int) mCommandData[TAN_SPEED]);
		AddOdometryData("ROTSPEED", (int) mCommandData[ROT_SPEED]);
		SendAllStringMessages();
			
		/*if(needodo)
		{
			printf("+++++++++++++++++++++++++++\n");
			ReadEncoders();
			ComputeOdometry();
			SendMotorMessage();
		}*/	
		
	}
	
};

void MotorExpert::RunClose()
{
  printf("\nMotorExpert closed");
  fflush(stdout);
};


void MotorExpert::SendMotorMessage()
{
	NewMessage(MSG_FROM_MOTION,MOTION_DATA);
	//AddOdometryData("TANSPEED", tanSpeed);
	//AddOdometryData("ROTSPEED", rotSpeed);
	SendAllStringMessages();
}

void MotorExpert::AddOdometryData(string name, float value)
{
  mCurrentMessage << "<" << value << ">";
}


long MotorExpert::elapsed(int i)
{ 
	struct timeval now;
	long elapsed_usec = 0;

	gettimeofday(&now, NULL);

	if (now.tv_usec > 1000000) {
		now.tv_usec -= 1000000;
		++now.tv_sec;
	}

	if (now.tv_sec > last_render[i].tv_sec){
		elapsed_usec = ((now.tv_sec - last_render[i].tv_sec) * 1000000);
	}

	if (now.tv_usec > last_render[i].tv_usec)
		elapsed_usec += now.tv_usec - last_render[i].tv_usec;
	else
		elapsed_usec -= last_render[i].tv_usec - now.tv_usec;


	return elapsed_usec ;

}

void MotorExpert::settimer(int i)
{
	gettimeofday(&(last_render[i]),NULL);
	return;
}
