
#include <stdio.h>
#include <stdlib.h>
#include <wiiuse.h>
#include <sys/time.h>
#include <math.h>
#include <unistd.h>

#include <string>
#include <sstream>

#include "wiim.h"

using namespace std;

#define MAX_WIIMOTES	1

#define GR
#define SC  2
#define PI 3.14159265358979323846

//#define CAMERALEFT
//#define CAMERARIGHT
#define GAMECONTROL

wiimote_t** wiimotes = NULL;
wiimote_t* wmLeft = NULL;
wiimote_t* wmRight = NULL;

wiimote_t* wm0 = NULL;

Wiim::Wiim()
{
	int found, connected;
	int i;
	
	fAuto=STARTMODE;
	BlinkRightStatus=0;
	BlinkLeftStatus=0;
	RumbleWStatus=0;
	RumbleSStatus=0;
	pointDiff=0;
	
	printf("\nwiim started\n"); 

	//Initialize an array of wiimote objects.
	wiimotes =  wiiuse_init(MAX_WIIMOTES);
	
	//Find wiimote devices, before to camera, after to control game
	
	printf ("\nPlease, connect left wiimote, right wiimote and gamecontrol");
	fflush(stdout);
	found = 0;
	while (!found) {
		found=wiiuse_find(wiimotes, MAX_WIIMOTES, 1);
	}

	//Connect to the wiimotes

	connected = wiiuse_connect(wiimotes, MAX_WIIMOTES);
	if (connected)
		printf("\nConnected to %i wiimotes (of %i found).", connected, found);
	else {
		printf("\nFailed to connect to any wiimote.");
	}

	/*#ifdef GAMECONTROL
	usleep(1000000);//stop blinking
	
	//wiiuse_rumble(wiimotes[0], 1);
	for(i=0;i<6;i++){
		wiiuse_set_leds(wiimotes[0], WIIMOTE_LED_1);
		usleep(50000);
		wiiuse_set_leds(wiimotes[0], WIIMOTE_LED_4);
		usleep(50000); 
	}
	for(i=0;i<6;i++){
		wiiuse_set_leds(wiimotes[1], WIIMOTE_LED_1);
		usleep(50000);
		wiiuse_set_leds(wiimotes[1], WIIMOTE_LED_4);
		usleep(50000); 
	}
	for(i=0;i<3;i++){
		wiiuse_set_leds(wiimotes[2], WIIMOTE_LED_1);
		usleep(50000); 
		wiiuse_set_leds(wiimotes[2], WIIMOTE_LED_2);
		usleep(50000); 
		wiiuse_set_leds(wiimotes[2], WIIMOTE_LED_3);
		usleep(50000); 
		wiiuse_set_leds(wiimotes[2], WIIMOTE_LED_4);
		usleep(50000); 
	}
	#endif*/
	
	#ifdef CAMERARIGHT
	//Set camera left
	usleep(1000000);
	wiiuse_set_leds(wiimotes[0], 4);
	wiiuse_rumble(wiimotes[0], 0);
	wiiuse_motion_sensing(wiimotes[0], 0);//enable accelerometer	
	usleep(10000);
	wiiuse_set_flags(wiimotes[0],WIIUSE_CONTINUOUS,WIIUSE_ORIENT_THRESH);
	usleep(10000);
	wiiuse_set_ir(wiimotes[0], 1);
	wiiuse_set_leds(wiimotes[0], WIIMOTE_LED_1);
	wmRight = wiimotes[0];
	#endif

	#ifdef CAMERALEFT
	//Set camera right
	usleep(1000000);
	wiiuse_set_leds(wiimotes[1], 2);
	wiiuse_rumble(wiimotes[1], 0);
	wiiuse_motion_sensing(wiimotes[1], 0);//enable accelerometer		
	usleep(10000);
	wiiuse_set_flags(wiimotes[1],WIIUSE_CONTINUOUS,WIIUSE_ORIENT_THRESH);
	usleep(10000);
	wiiuse_set_ir(wiimotes[1], 1);
	wiiuse_set_leds(wiimotes[1], WIIMOTE_LED_2);
	wmLeft = wiimotes[1];
	#endif
	
	#ifdef GAMECONTROL
	//Set game control
	usleep(1000000);
	wiiuse_set_leds(wiimotes[0], 1);
	wiiuse_rumble(wiimotes[0], 0);
	wiiuse_motion_sensing(wiimotes[0], 1);//enable accelerometer		
	usleep(10000);
	wiiuse_set_flags(wiimotes[0],WIIUSE_CONTINUOUS,WIIUSE_ORIENT_THRESH);
	usleep(10000);
	wiiuse_set_ir(wiimotes[0], 1);
	wm0 = wiimotes[0];
	if(fAuto==0){
		wiiuse_set_leds(wm0, WIIMOTE_LED_1 | WIIMOTE_LED_4);
	}
	#endif
}

Wiim::~Wiim()
{
}

int Wiim::run()
{
	
	int i=0;
		
		if(!wiiuse_poll(wiimotes, MAX_WIIMOTES)){
			return 0;
		}else{
			BlinkRight(BlinkRightStatus);
			BlinkLeft(BlinkLeftStatus);
			WeakRumble(RumbleWStatus);
			StrongRumble(RumbleSStatus);
			return 1;
		}
	
	
}



void Wiim::handle_ctrl_status(struct wiimote_t* wm){
	printf("get status\n");
	printf("\n--- CONTROLLER STATUS [wiimote id %i] ---", wm->unid);
	printf("attachment:      %i\n", wm->exp.type);
	printf("speaker:         %i\n", WIIUSE_USING_SPEAKER(wm));
	printf("ir:              %i\n", WIIUSE_USING_IR(wm));
	printf("leds:            %i %i %i %i\n", WIIUSE_IS_LED_SET(wm, 1), WIIUSE_IS_LED_SET(wm, 2), WIIUSE_IS_LED_SET(wm, 3), WIIUSE_IS_LED_SET(wm, 4));
	printf("battery:         %f %%\n", wm->battery_level);
}

void Wiim::handle_event(struct wiimote_t* wm) {
	
	int i;
			
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_A))		printf("A pressed\n");
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_B))		printf("B pressed\n");
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_UP))		printf("UP pressed\n");
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_DOWN))	printf("DOWN pressed\n");
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_LEFT))	printf("LEFT pressed\n");
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_RIGHT))	printf("RIGHT pressed\n");
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_MINUS))	printf("MINUS pressed\n");
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_PLUS))	printf("PLUS pressed\n");
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_ONE))		printf("ONE pressed\n");
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_TWO))		printf("TWO pressed\n");
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_HOME))	printf("HOME pressed\n");
			
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_A)){
				printf("accellerometer enabled\n");
				wiiuse_motion_sensing(wm, 1);//enable accelerometer				
			}
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_B)){
				printf("accellerometer disabled\n");
				wiiuse_motion_sensing(wm, 0);//disable accelerometer
			}
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_PLUS)){
				printf("ir enabled\n");
				wiiuse_set_ir(wm, 1);//enable ir
			}
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_MINUS)){
				printf("ir disabled\n");
				wiiuse_set_ir(wm, 0);//disable ir
			}
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_HOME)){
				printf("status requested\n");
				wiiuse_status(wm);//request status
			}
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_ONE)){
				printf("position tracking enabled\n");
				wiiuse_motion_sensing(wm, 1);
			}
			if (IS_PRESSED(wm, WIIMOTE_BUTTON_TWO)){
				printf("position tracking disabled\n");
				wiiuse_motion_sensing(wm, 0);
			}
			
			if (WIIUSE_USING_ACC(wm)) {
				printf("wiimote calibration xyz 0g: %i %i %i --- 1g: %i %i %i\t%i\n",
						wm->accel_calib.cal_zero.x,wm->accel_calib.cal_zero.y,wm->accel_calib.cal_zero.z,
						wm->accel_calib.cal_g.x,wm->accel_calib.cal_g.y,wm->accel_calib.cal_g.z);
	      printf("x:%f y:%f z:%f \n",wm->gforce.x,wm->gforce.y,wm->gforce.z);
			}

			if (WIIUSE_USING_IR(wm)) {
				// definisce il numero dei led da rilevare 
				for (i=0; i < 3; i++) {
					if (wm->ir.dot[i].visible){
						printf("IR%i ---: (%u, %u) %u\t\t", i, wm->ir.dot[i].x, wm->ir.dot[i].y,wm->ir.dot[i].size);
					}else{
						printf("IR%i ***: (%u, %u) %u\t\t", i, wm->ir.dot[i].x, wm->ir.dot[i].y,wm->ir.dot[i].size);
					}
				}
				printf("\n");
			}
}

/*
float Wiim::get_X_acceleration() {
	return X_acceleration;
}

float Wiim::get_Y_acceleration() {
	return Y_acceleration;
}

float Wiim::get_Z_acceleration() {
	return Z_acceleration;
}
*/

void Wiim::SetpointDiff(int pd)
{
	pointDiff=pd;
}

void Wiim::BlinkRight(int status)
{
	if (status==0){
		return;
	}
	if(status<239){
		if ((status%32)>=0 && (status%32)<8){
			wiiuse_set_leds(wm0, WIIMOTE_LED_1);
		}
		if ((status%32)>8 && (status%32)<16){
			wiiuse_set_leds(wm0, WIIMOTE_LED_2);
		}
		if ((status%32)>16 && (status%32)<24){
			wiiuse_set_leds(wm0, WIIMOTE_LED_3);
		}
		if ((status%32)>24 && (status%32)<32){
			wiiuse_set_leds(wm0, WIIMOTE_LED_4);
		}
	}
	status++;
	if(status==240){
		showPoint();
	}
	if(status>550){
		status=0;
		wiiuse_set_leds(wm0, 0);
	}
	BlinkRightStatus=status;
	return;
}

void Wiim::BlinkLeft(int status)
{
	if (status==0){
		return;
	}
	if(status<239){
		if ((status%32)>=0 && (status%32)<8){
			wiiuse_set_leds(wm0, WIIMOTE_LED_4);
		}
		if ((status%32)>8 && (status%32)<16){
			wiiuse_set_leds(wm0, WIIMOTE_LED_3);
		}
		if ((status%32)>16 && (status%32)<24){
			wiiuse_set_leds(wm0, WIIMOTE_LED_2);
		}
		if ((status%32)>24 && (status%32)<32){
			wiiuse_set_leds(wm0, WIIMOTE_LED_1);
		}
	}
	status++;
	if(status==240){
		showPoint();
	}
	if(status>550){
		status=0;
		wiiuse_set_leds(wm0, 0);
	}
	BlinkLeftStatus=status;
	return;
}

void Wiim::WeakRumble(int status){
	if (status==0){
		return;
	}
	if (status<15){
		wiiuse_rumble(wm0, 1);
		status++;
	}else{
		wiiuse_rumble(wm0, 0);
		status=0;
	}
	RumbleWStatus=status;
}

void Wiim::StrongRumble(int status){
	if (status==0){
		return;
	}
	if (status<150){
		wiiuse_rumble(wm0, 1);
		status++;
	}else{
		wiiuse_rumble(wm0, 0);
		status=0;
	}
	RumbleSStatus=status;
}

int Wiim::IsAuto()
{
	if (IS_PRESSED(wm0, WIIMOTE_BUTTON_PLUS)){ 
		fAuto=1;
		wiiuse_set_ir(wm0, 1);
		wiiuse_set_leds(wm0, 0);
	}
	if (IS_PRESSED(wm0, WIIMOTE_BUTTON_MINUS)){ 
		fAuto=0;
		showPoint();
	}

	if(fAuto){
		return 1;
	}else{
		return 0;
	}
}

void Wiim::showPoint()
{	
	switch (pointDiff){
		case 0:
			wiiuse_set_leds(wm0, WIIMOTE_LED_2 | WIIMOTE_LED_3);
		break;
		case 1:
			wiiuse_set_leds(wm0,WIIMOTE_LED_3);
		break;
		case 2:
			wiiuse_set_leds(wm0, WIIMOTE_LED_3 | WIIMOTE_LED_4);
		break;
		case 3:
			wiiuse_set_leds(wm0, WIIMOTE_LED_4);
		break;
		case 4:
			wiiuse_set_leds(wm0,  WIIMOTE_LED_2 | WIIMOTE_LED_3 | WIIMOTE_LED_4);
		break;
		case -1:
			wiiuse_set_leds(wm0, WIIMOTE_LED_2 );
		break;
		case -2:
			wiiuse_set_leds(wm0, WIIMOTE_LED_1 | WIIMOTE_LED_2);
		break;
		case -3:
			wiiuse_set_leds(wm0, WIIMOTE_LED_1);
		break;
		case -4:
			wiiuse_set_leds(wm0,  WIIMOTE_LED_1 | WIIMOTE_LED_2 | WIIMOTE_LED_3);
		break;
		default:
			wiiuse_set_leds(wm0,  WIIMOTE_LED_1 | WIIMOTE_LED_2 | WIIMOTE_LED_3 | WIIMOTE_LED_4);
		break;
	}
}
		
void Wiim::SetCharge(float c)
{
	if(BlinkRightStatus!=0 || BlinkLeftStatus!=0){
		return;
	}
	if(c<1){
		wiiuse_set_leds(wm0, 0);
	}
	if(c>=1 && c<2){
		wiiuse_set_leds(wm0, WIIMOTE_LED_1);
	}else if(c>=2 && c<3){
		wiiuse_set_leds(wm0, WIIMOTE_LED_1 |  WIIMOTE_LED_2);
	}else if(c>=3 && c<4){
		wiiuse_set_leds(wm0, WIIMOTE_LED_1 |  WIIMOTE_LED_2 | WIIMOTE_LED_3);
	}else if(c>=4){
		wiiuse_set_leds(wm0, WIIMOTE_LED_1 |  WIIMOTE_LED_2 | WIIMOTE_LED_3 | WIIMOTE_LED_4);
	}
}	
		

float Wiim::getPitch()
{
	float p;
	p=wm0->orient.pitch;
	//printf("%f\n",p);
	return p;
}



string Wiim::getToMotor()
{
	string buf;
	//controllo manuale
	if (IS_PRESSED(wm0, WIIMOTE_BUTTON_ONE)){ //avanti
		buf += "TanSpeed";
		buf += " 0.0";
		return buf;
	}
	if (IS_PRESSED(wm0, WIIMOTE_BUTTON_UP)){ //avanti
		buf += "TanSpeed";
		buf += " 50.0";
		return buf;
	}
	if (IS_PRESSED(wm0, WIIMOTE_BUTTON_DOWN)){ //indietro
		buf += "TanSpeed";
		buf += " -50.0";
		return buf;
	}
	if (IS_PRESSED(wm0, WIIMOTE_BUTTON_RIGHT)){ //destra 
		buf += "RotSpeed";
		buf += " -50.0";
		return buf;
	}
	if (IS_PRESSED(wm0, WIIMOTE_BUTTON_LEFT)){ //sinistra
		buf += "RotSpeed";
		buf += " +50.0";
		return buf;
	}
	
	return "";
}

int Wiim::getEmergencyStop()
{
	if (IS_PRESSED(wm0, WIIMOTE_BUTTON_TWO)){
		return 1;
	}else{
		return 0;
	}
}

int Wiim::getButtonB()
{
	if (IS_PRESSED(wm0, WIIMOTE_BUTTON_B)) { 
		return 1;
	}
	return 0;
}

int Wiim::getJButtonHome()
{
	if (IS_JUST_PRESSED(wm0, WIIMOTE_BUTTON_HOME)){
		return 1;
	}else{
		return 0;
	}
}

int Wiim::getHButtonA()
{
	if (IS_HELD(wm0, WIIMOTE_BUTTON_A)){
		return 1;
	}else{
		return 0;
	}
}
	
float Wiim::GetAcceleration()
{
	float ta;
	ta = sqrt(pow(wm0->gforce.x,2)+pow(wm0->gforce.y,2)+pow(wm0->gforce.z,2));
	//printf (" -- %f\n",ta);
	return ta;
}

ppoint Wiim::getcenter()
{

	int x=0,y=0;
	ppoint p;
	
	//printf("% d\t%d\t%d\n",wm0->ir.num_dots,wm0->ir.ax,wm0->ir.ay);
	if (wm0->ir.num_dots>0){
		if(wm0->ir.ax>=0 && wm0->ir.ax <=1024 && wm0->ir.ay>=0 && wm0->ir.ay <=768){
			p.x=wm0->ir.ax;
			p.y=wm0->ir.ay;
			p.valid=1;
		}else{
			p.valid=0;
		}
	}else{
		p.valid=0;
	}
	return p;
}

int Wiim::getpoints(lpoint *p)
{

	int x=0, y=0, pointsNumber = 0, distMax = -1, m1 = -1, m2 = -1, yMax = 0;

	//p = (lpoint*) malloc(sizeof(lpoint) * 4);
	
	#ifdef CAMERARIGHT
	for(int i = 0; i < 3; i++)
	{
		for(int c = i + 1; c < 4; c++)
		{
			if(wmRight->ir.dot[i].visible && wmRight->ir.dot[i].rx >= 0 && wmRight->ir.dot[i].rx <= 1024 && wmRight->ir.dot[i].ry >= 0 && wmRight->ir.dot[i].ry <= 768 && wmRight->ir.dot[c].visible && wmRight->ir.dot[c].rx >= 0 && wmRight->ir.dot[c].rx <= 1024 && wmRight->ir.dot[c].ry >= 0 && wmRight->ir.dot[c].ry <= 768)
			{
				if (yMax < wmRight->ir.dot[i].ry)
					yMax = wmRight->ir.dot[i].ry;
				if (((wmRight->ir.dot[i].rx - wmRight->ir.dot[c].rx)*(wmRight->ir.dot[i].rx - wmRight->ir.dot[c].rx)) > distMax)
				{
					distMax = (wmRight->ir.dot[i].rx - wmRight->ir.dot[c].rx)*(wmRight->ir.dot[i].rx - wmRight->ir.dot[c].rx);
					m1 = i;
					m2 = c;
				}
			}
		}
	}
	
	if ((m1 != -1) && (m2 != -1))
	{
		p[0].x=wmRight->ir.dot[m1].rx;
		p[0].y=wmRight->ir.dot[m1].ry;
		p[0].type=0;
		p[1].x=wmRight->ir.dot[m2].rx;
		p[1].y=wmRight->ir.dot[m2].ry;
		p[1].type=0;
		if (p[0].x < p[1].x)
		{
			int xTemp = p[0].x;
			p[0].x = p[1].x;
			p[1].x = xTemp;
		}
		
		//printf("\nXY: %d, %d", p[0].x, p[0].y);
		//printf("\nXY: %d, %d", p[1].x, p[1].y);
		pointsNumber += 2;
	}
	#endif

	#ifdef CAMERALEFT
	distMax = -1;
	m1 = -1;
	m2 = -1;
	for(int i = 0; i < 3; i++)
	{
		for(int c = i + 1; c < 4; c++)
		{
			if(wmLeft->ir.dot[i].visible && wmLeft->ir.dot[i].rx >= 0 && wmLeft->ir.dot[i].rx <= 1024 && wmLeft->ir.dot[i].ry >= 0 && wmLeft->ir.dot[i].ry <= 768 && wmLeft->ir.dot[c].visible && wmLeft->ir.dot[c].rx >= 0 && wmLeft->ir.dot[c].rx <= 1024 && wmLeft->ir.dot[c].ry >= 0 && wmLeft->ir.dot[c].ry <= 768)
			{
				if (((wmLeft->ir.dot[i].rx - wmLeft->ir.dot[c].rx)*(wmLeft->ir.dot[i].rx - wmLeft->ir.dot[c].rx)) > distMax)
				{
					distMax = (wmLeft->ir.dot[i].rx - wmLeft->ir.dot[c].rx)*(wmLeft->ir.dot[i].rx - wmLeft->ir.dot[c].rx);
					m1 = i;
					m2 = c;
				}
			}
		}
	}
	if ((m1 != -1) && (m2 != -1))
	{
		p[2].x=wmLeft->ir.dot[m1].rx;
		p[2].y=wmLeft->ir.dot[m1].ry;
		p[2].type=0;
		p[3].x=wmLeft->ir.dot[m2].rx;
		p[3].y=wmLeft->ir.dot[m2].ry;
		p[3].type=0;
		if (p[2].x < p[3].x)
		{
			int xTemp = p[2].x;
			p[2].x = p[3].x;
			p[3].x = xTemp;
		}
		//printf("\nXY: %d, %d", p[2].x, p[2].y);
		//printf("\nXY: %d, %d", p[3].x, p[3].y);
		pointsNumber += 20;
	};
	#endif

	#ifdef GAMECONTROL
	/*#ifndef CAMERARIGHT
	#ifndef CAMERALEFT
	wm = wiimotes[0];
	#else
	wm = wiimotes[1];
	#endif
	#else
	#ifndef CAMERALEFT
	wm = wiimotes[1];
	#else
	wm = wiimotes[2];
	#endif
	#endif*/
	/*for(int i = 0; i < wm0->ir.num_dots; i++)
	{
		if(wm0->ir.dot[i].visible && wm0->ir.dot[i].rx >= 0 && wm0->ir.dot[i].rx <= 1024 && wm0->ir.dot[i].ry >= 0 && wm0->ir.dot[i].ry <= 768)
		{
			current = (lpoint*) malloc(sizeof(lpoint));
			if (!p)
			{
				p = current;
				end = p;
			}
			else
			{
				end->next = current;
				end = current;
			}
			current->x=wm0->ir.dot[i].rx;
			current->y=wm0->ir.dot[i].ry;
			current->type=2;
			current->next = NULL;
		}
	}*/
	#endif

	return pointsNumber;
}
