/*
 * AIRboardUI: graphical interface program to interact with airboard
 * through RS232 serial port.
 *
 * In order to compile this package you have to install the
 * Java Communications API on your system.
 * http://java.sun.com/products/javacomm/downloads/index.html
 * 
 * Bruschieri Stefano
 * Campari PierPaolo
 */

package airboardUI;
import java.awt.Button;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Label;
import java.awt.Panel;
import java.awt.TextField;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.InputStream;
import java.io.OutputStream;
import javax.comm.*;
import javax.swing.Timer;

import java.util.Vector;

/**
 * @author Bruschieri Stefano
 */
public class FuzzyForm extends Frame implements ActionListener {
		
	private static final int INCR_MAX = 0;
	private static final int SMALL = 1;
	private static final int HIGH = 2;
	
    private int MAX_ERRORS = 5;
    private int MAX_NO_RESPONSE = 5;
    
	private static final int WIDTH = 300;
	private static final int HEIGHT = 150;
    private SerialConnection connection;
	private AirboardFrame parent;
    
	private Panel formPanel;
    private Button writeButton;
    private TextField inputOneTextField;
    private TextField inputTwoTextField;
    private TextField inputThreeTextField;
	
    private Vector commands;
  
    ProgressDialog pd;
    Timer timer;
    int i;
	int noResponseCounter = 0;
	int errorsCounter = 0;
    
	public FuzzyForm(SerialConnection connection, AirboardFrame parent){
		super();
		this.connection = connection;
		this.parent = parent;
		addWindowListener(new CloseHandler(this));	
		
		formPanel = new Panel();
		writeButton = new Button("Write");
		writeButton.addActionListener(this);
		inputOneTextField = new TextField(20);
		inputOneTextField.setEditable(true);
		inputTwoTextField = new TextField(20);
		inputTwoTextField.setEditable(true);
		inputThreeTextField = new TextField(20);
		inputThreeTextField.setEditable(true);

		GridBagLayout gridBag = new GridBagLayout();
		GridBagConstraints cons = new GridBagConstraints();
		cons.gridwidth = GridBagConstraints.REMAINDER;
		cons.weightx = 1.0;
		formPanel.setLayout(new GridLayout(4,2));
		formPanel.add(new Label("Max increment:"));
		formPanel.add(inputOneTextField);
		formPanel.add(new Label("Pos Small:"));
		formPanel.add(inputTwoTextField);
		formPanel.add(new Label("Pos High:"));
		formPanel.add(inputThreeTextField);
		gridBag.setConstraints(formPanel, cons);
		add(formPanel);
		add(writeButton, "Center");
		this.setLayout(gridBag);
		
		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
		this.setLocation(screenSize.width/2 - WIDTH/2, 
			    screenSize.height/2 - HEIGHT/2);
		this.setSize(WIDTH, HEIGHT);
		this.setVisible(true);
		this.repaint();	
	}
		
	/**
	 * The specified String <code>text</code> is converted to an integer, if
	 * this number is out of allowed range or its format is not correct, this
	 * function simply return a string with no lenght.
	 * 
	 * @param text The value of parameter 
	 * @param type The type of command defined in this class
	 * @return The command to send to the PIC
	 */
	private String prepareString(String text, int type){
		String toReturn = "";
		
    	/* integer value of text must be >0 and <255*/
		try {
			if( (new Integer(text)).intValue()<0 || 
					(new Integer(text)).intValue()>255 ){
				System.out.println("Errore: value out of range "+Integer.valueOf(text));
				return "";
			}
		} catch(NumberFormatException nfe){
			return "";
		}
    	
    	if(text.length()==1) text = "00"+text;
    	if(text.length()==2) text = "0"+text;
    	if(text.length()==0 || text.length()>3) 
    		text="000";
    	
    	if( type == FuzzyForm.INCR_MAX )
    		toReturn = "ewfincrmax=+"+text+"&+000";
    	if( type == FuzzyForm.SMALL )
    		toReturn = "ewfsmalout=+"+text+"&+000";
    	if( type == FuzzyForm.HIGH )
    		toReturn = "ewfhighout=+"+text+"&+000";
		return toReturn;
	}
	
    /**
    * Responds to the buttons.
    */
    public void actionPerformed(ActionEvent e) {
    	String cmd = e.getActionCommand();
    
    	if( cmd.equals("Write") ){
    		this.commands = new Vector();
    		String s="";
    		if( this.inputOneTextField.getText().length()>0 ){
    			s = this.prepareString(this.inputOneTextField.getText(), FuzzyForm.INCR_MAX);
    			if( s.length()>0 )
    				this.commands.add(s);
    		}

    		if( this.inputTwoTextField.getText().length()>0 ){
    			s = this.prepareString(this.inputTwoTextField.getText(), FuzzyForm.SMALL);
    			if( s.length()>0 )
    				this.commands.add(s);
    		}
    		
    		if( this.inputThreeTextField.getText().length()>0 ){
    			s = this.prepareString(this.inputThreeTextField.getText(), FuzzyForm.HIGH);
    			if( s.length()>0 )
    				this.commands.add(s);
    		}
    		this.setVisible(false);
    		System.out.println(this.commands.size());
    		this.transmitChanges();
    	}
    }
	   
    /**
     *  Sends one command each time through serial port. All of changes are stored
     *  in a vector of Strings and, when user clicks on Apply button, AIRboardUI
     *  reads runs over the vector and, after waiting for the DELAY_TIME, sends
     *  the commands it reads. 
     */
    public void transmitChanges(){    	
    	pd = new ProgressDialog(this.parent, this.commands.size() );
		
		SerialPort sPort = this.connection.getSerialPort();
		OutputStream os;

    	/* timer */
    	timer = new Timer(AirboardFrame.DELAY_TIME, new ActionListener() {		
    		public void actionPerformed(ActionEvent evt) {
    		SerialPort sPort;
    		OutputStream os;
    		InputStream is;
    	    	
    		if(!pd.hasBeenStopped() && 
    				i<=( commands.size()-1)){
    					pd.progressBar.setValue(i);
    			    	pd.setText("Appling changes...");
    	
    					String lastInput = (connection.getLastInputString());
    					
    					if(lastInput.length()>0){
    						if(lastInput.lastIndexOf("ERROR")!=-1 ){
    							System.out.println(i+" An error was occurred ");
        						if(errorsCounter >= MAX_ERRORS) { 
        							pd.setCancelled("AIRLab returns error"); 
        							pd.setCompleted(); 
        						}
    							return;
    						} else { 
    							System.out.println(i+" OK"); 
    							i++; 
    						}    			
    					} else {
    						noResponseCounter++;
    						System.out.println("No response");
    						if(noResponseCounter>=MAX_NO_RESPONSE) { 
    							pd.setCancelled("No response"); 
    							pd.setCompleted(); 
    						}
    						return;
    					} 

    					String toOutput;
    					toOutput = (commands.elementAt((i-1))).toString();

    					connection.sendAString(toOutput);
    					System.out.println(toOutput);
    				} else { 
    					pd.setCompleted();
    					timer.stop(); 
    				}}
    		});

    		timer.start();
    }
    
    /**
    * Handles closing down system. Allows application to be closed with window
    * close box.
    */
    class CloseHandler extends WindowAdapter {

	FuzzyForm ff;
	
	public CloseHandler(FuzzyForm ff) {
	    this.ff = ff;
	}

	public void windowClosing(WindowEvent e) {
		ff.setVisible(false);
	}
    }
}
